# Copyright (c) 2024 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._builder import (
    UICR_MAGIC,
    UICR_VERSIONS,
    UICR_VERSION_MIN,
    UICR_VERSION_MAX,
    Uicr,
    UicrError,
    UicrVersion,
    Header,
    Memory,
    Mailbox,
    Peripheral,
    Pin,
    Feature,
    DppiLink,
    IpcLink,
    Trace,
    TracePortConfig,
    EtrBuffer,
)
from ._common import (
    UnitSrc,
    Value,
)

__all__ = [
    "UICR_MAGIC",
    "UICR_VERSIONS",
    "UICR_VERSION_MIN",
    "UICR_VERSION_MAX",
    "Uicr",
    "UicrError",
    "UicrVersion",
    "Header",
    "Memory",
    "Mailbox",
    "Peripheral",
    "Pin",
    "Feature",
    "DppiLink",
    "IpcLink",
    "Trace",
    "TracePortConfig",
    "EtrBuffer",
    "UnitSrc",
    "Value",
]



# Copyright (c) 2025 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

import svd

from nrfregtool._svd_util import content_truncate_and_set


UnitSrc = str | list[str]
"""A generic specifier for a source location.
This can be given to some APIs to attach a context to a configuration.
The context can later be used to give better information and error messages.
"""


T = TypeVar("T")


@dataclass(frozen=True)
class Value(Generic[T]):
    """A value with an associated source context.

    These values can be passed to the UICR API to associate debug information
    with the configuration values.

    :var val: the value.
    :var src: arbitrary source context string(s).
    """

    val: T
    src: UnitSrc


class RegSrcMap:
    def __init__(self) -> None:
        self._sources = defaultdict(list)

    def reg_set(
        self,
        reg: svd.Register | svd.Field,
        val: str | int,
        src: UnitSrc | svd.Register | svd.Field | None,
        src_mode: Literal["append", "overwrite"] = "append",
        lsb_truncate: bool = False,
    ) -> None:
        if lsb_truncate:
            assert isinstance(reg, svd.Field)
            content_truncate_and_set(reg, val)
        else:
            reg.content = val

        match src:
            case svd.Register() | svd.Field():
                new_src = self._sources[reg_path_make(src)]
            case list():
                new_src = src
            case str():
                new_src = [src]
            case _:
                new_src = []

        if src_mode == "overwrite":
            self._sources[reg_path_make(reg)] = list(new_src)
        elif src_mode == "append":
            existing = self._sources[reg_path_make(reg)]
            for s in new_src:
                if s not in existing:
                    existing.append(s)
        else:
            raise ValueError(f"Invalid {src_mode=}")

    def reg_clear(self, reg: svd.Register | svd.Field) -> None:
        reg.content = reg.reset_content
        reg_path = reg_path_make(reg)
        for key, sources in self._sources.items():
            if key.startswith(reg_path):
                sources.clear()

    def reg_move(
        self,
        reg_to: svd.RegisterUnion | svd.Field,
        reg_from: svd.RegisterUnion | svd.Field,
    ) -> None:
        src = self._sources.pop(reg_path_make(reg_from), [])
        if src:
            self._sources[reg_path_make(reg_to)].extend(src)

        if isinstance(reg_from, svd.Field):
            reg_to.content = reg_from.content
            reg_from.content = reg_from.reset_content
            return
        for child in reg_from:
            self.reg_move(reg_to[child], reg_from[child])

    @property
    def src_dict(self) -> dict[str, list[str]]:
        return self._sources

    def src_lookup(self, reg: svd.RegisterUnion) -> list[str]:
        path = reg_path_make(reg)
        combined = []
        for key, sources in self._sources.items():
            if key.startswith(path):
                combined.extend(sources)
        return combined


def reg_path_make(reg: svd.RegisterUnion | svd.Field) -> str:
    match reg:
        case svd.Array() | svd.Struct() | svd.Register():
            return repr(reg.path)
        case svd.Field():
            return f"{reg._register.path!r}:{reg.name}"
        case _:
            raise ValueError(f"Unsupported reg: {reg}")



# Copyright (c) 2024 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import enum
import inspect
import pickle
import re
from dataclasses import dataclass
from functools import wraps
from itertools import islice
from pathlib import Path
from typing import (
    Any,
    Callable,
    Optional,
    Iterable,
    Iterator,
    Type,
    TypeVar,
    Sequence,
    overload,
)

from intelhex import IntelHex

from devicetree import edtlib

from nrfregtool import log, uicr
from nrfregtool.platform import (
    CTRLSEL_DEFAULT,
    AddressRegion,
    Ctrlsel,
    ProcessorID,
    Product,
    DomainID,
    OwnerID,
    secure_address_get,
)

NON_PROG_OWNER_COMPATS = [
    "nordic,nrf-gpio",
    "nordic,nrf-gpiote",
    "nordic,nrf-grtc",
    "nordic,nrf-dppic-global",
    "nordic,nrf-ipct-global",
    "nordic,nrf-clic",
    "nordic,nrf-bellboard",
    "nordic,nrf-vevif-task-tx",
    # Note: this has programmable ownership, but this aligns with current usage of the peripheral
    # where the secure domain configures it.
    "nordic,nrf-tddconf",
]
NON_PROG_OWNER_COMPAT_PATTERN = (
    "^" + "|".join((f"({c})" for c in NON_PROG_OWNER_COMPATS)) + "$"
)
# Every compatible starting with "nordic," that is not a non-programmable ownership compatible
PROG_OWNER_COMPAT_PATTERN = rf"(?!{NON_PROG_OWNER_COMPAT_PATTERN})^nordic,.+$"
# Status values that indicate ownership either by the processor or a child processor
STATUS_OWNED = ("okay", "reserved")


class BadDevicetreeError(RuntimeError):
    """Error raised when the devicetree is misconfigured
    or if assumptions about the devicetree are not met.
    """

    ...


C = TypeVar("C")
VisitMethod = Callable[[C, edtlib.Node], None]
BoundVisitMethod = Callable[[edtlib.Node], None]


@dataclass
class DtMatch:
    """Decorator used to annotate DT node visitor methods with a devicetree node pattern.

    A node match occurs when all of the parameters in the pattern match the node.

    :param compat: node compatible string or regex, matched against every member in node.compats.
    :param path: node path string or regex, matched against node.path.
    :param label: node label string or regex, matched against every member in node.labels.
    :param status: one or more node status strings, matched against node.status.
    :param props: list of property name string or regex, matched against node.props.keys().
    """

    compat: str | re.Pattern | None = None
    path: str | re.Pattern | None = None
    label: str | re.Pattern | None = None
    status: str | Sequence[str] | None = None
    regs: Callable[[int], bool] | None = None
    props: Sequence[str | re.Pattern] | None = None

    def __call__(self, method: VisitMethod) -> VisitMethod:
        @wraps(method)
        def visit_method_wrapper(c: C, node: edtlib.Node, /) -> None:
            match self.path:
                case str(path):
                    if node.path != path:
                        return
                case re.Pattern() as pattern:
                    if not re.search(pattern, node.path):
                        return

            match self.compat:
                case str(compat):
                    if not any((c == compat for c in node.compats)):
                        return
                case re.Pattern() as pattern:
                    if not any((re.search(pattern, c) for c in node.compats)):
                        return

            match self.label:
                case str(label):
                    if not any((l == label for l in node.labels)):
                        return
                case re.Pattern() as pattern:
                    if not any((re.search(pattern, l) for l in node.labels)):
                        return

            match self.status:
                case str(status):
                    if node.status != status:
                        return
                case list() | tuple() as statuses:
                    if node.status not in statuses:
                        return

            if isinstance(self.regs, Callable):
                if not node.regs or not self.regs(node.regs[0].addr):
                    return

            if self.props:
                any_found = False

                for expected in self.props:
                    match expected:
                        case str(prop):
                            any_found = prop in node.props

                        case re.Pattern() as pattern:
                            any_found = any((re.search(pattern, p) for p in node.props))

                    if any_found:
                        break

                if not any_found:
                    return

            method(c, node)

        return visit_method_wrapper


DtLocation = (
    edtlib.Node
    | edtlib.Property
    | tuple[edtlib.Property, int]
    | list[edtlib.Node | edtlib.Property | tuple[edtlib.Property, int]]
)
"""Specifer for a location in the devicetree."""


def uicr_from_devicetree(
    edt_pickle_file: Path,
    product: Product,
    output_file: Path,
    output_debug_file: Optional[Path] = None,
) -> None:
    """Populate a UICR based on a devicetree."""
    with edt_pickle_file.open("rb") as file:
        dt = pickle.load(file)

    processor = dt_processor_id(dt)
    uicr_builder = uicr.Uicr()

    visitor = UicrMatcher(builder=uicr_builder, product=product, processor=processor)
    visit_methods: list[BoundVisitMethod] = [
        method
        for name, method in inspect.getmembers(visitor, inspect.ismethod)
        if not name.startswith("_")
    ]
    for node in dt.nodes:
        for visit in visit_methods:
            visit(node)

    log.info(uicr_builder.pretty_print(verbose=log.getEffectiveLevel() < log.INFO))

    uicr_hex = IntelHex()
    uicr_hex.frombytes(uicr_builder.build_bytes(), offset=visitor.uicr_address)

    with output_file.open("w", encoding="utf-8") as file:
        uicr_hex.write_hex_file(file)
    if output_debug_file:
        with output_debug_file.open("w", encoding="utf-8") as file:
            file.write(uicr_builder.build_debug_info())


def address_is_global_periph(address: int) -> bool:
    try:
        domain = DomainID.from_address(address)
    except ValueError:
        return False

    if domain != DomainID.GLOBAL:
        return False

    if AddressRegion.from_address(address) not in (
        AddressRegion.PERIPHERAL,
        AddressRegion.STM,
    ):
        return False

    return True


class UicrMatcher:
    def __init__(
        self, builder: uicr.Uicr, product: Product, processor: ProcessorID
    ) -> None:
        self.uicr_address = 0
        self._processor = processor
        self._product = product
        self._builder = builder
        try:
            self._owner = OwnerID.from_processor(self._processor)
        except Exception:
            raise RuntimeError(
                f"Processor ID {self._processor} does not correspond to an owner"
            )
        try:
            self._domain = DomainID.from_processor(self._processor)
        except Exception:
            raise RuntimeError(
                f"Processor ID {self._processor} does not correspond to a domain"
            )

        header = uicr.Header(
            part_code=self._product.product_code.part_number,
            hardware_revision=self._product.product_code.revision,
            domain=self._domain.value,
        )
        self._builder.set_header(header)

    @DtMatch(path="/chosen")
    def visit_chosen(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        if prop := node._node.props.get("zephyr,code-partition"):
            code_partition = node.edt.get_node(prop.to_path().path)
            vtor_address = dt_partition_address_get(code_partition)
            secure = dt_mem_node_is_secure(code_partition)
            if secure.val:
                self._builder.set_secure_vtor(vtor_address)
            else:
                self._builder.set_nonsecure_vtor(vtor_address)

    @DtMatch(compat="nordic,owned-partitions", status=STATUS_OWNED)
    def visit_owned_partitions(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        if not node.children:
            return

        partitions = list(node.children.values())
        abs_addresses = [dt_partition_address_get(p) for p in partitions]
        end_addresses = [
            uicr.Value(a.val + p.regs[0].size, a.src)
            for a, p in zip(abs_addresses, partitions)
        ]
        min_addr = min(abs_addresses, key=lambda a: a.val)
        max_addr = max(end_addresses, key=lambda a: a.val)
        size = uicr.Value(max_addr.val - min_addr.val, src=max_addr.src)
        access = self._parse_access_props(node)

        memory = uicr.Memory(address=min_addr, size=size, access=access)
        self._builder.add_memory(memory)

    @DtMatch(compat="nordic,owned-memory", status=STATUS_OWNED)
    def visit_owned_memory(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        addr = dt_reg_addr(node)
        size = dt_reg_size(node)
        access = self._parse_access_props(node)

        memory = uicr.Memory(address=addr, size=size, access=access)
        self._builder.add_memory(memory)

    def _parse_access_props(self, node: edtlib.Node) -> list[uicr.Memory.Access]:
        accesses = []
        props = node.props

        if "nordic,access" not in props:
            # TODO: drop deprecated props with zephyr 4.2.0
            owner = dt_prop(node, "owner-id", default=self._owner.value)
            readable = dt_prop(node, "perm-read")
            writable = dt_prop(node, "perm-write")
            executable = dt_prop(node, "perm-execute")
            secure = dt_prop(node, "perm-secure")
            non_secure_callable = dt_prop(node, "non-secure-callable")

            access = uicr.Memory.Access(
                owner=owner,
                read=readable,
                write=writable,
                execute=executable,
                secure=secure,
                non_secure_callable=non_secure_callable,
            )
            accesses.append(access)
        else:
            if len(props["nordic,access"].val) % 2 != 0:
                raise BadDevicetreeError(
                    f"nordic,access property in {node.path} "
                    "does not contain an even number of values"
                )

            nordic_access = dt_array_prop(node, "nordic,access")
            for owner, permission in batched(nordic_access, 2):
                access_perm = NordicAccessPerm(permission.val)
                access = uicr.Memory.Access(
                    owner=owner,
                    read=uicr.Value(
                        NordicAccessPerm.NRF_PERM_R in access_perm, permission.src
                    ),
                    write=uicr.Value(
                        NordicAccessPerm.NRF_PERM_W in access_perm, permission.src
                    ),
                    execute=uicr.Value(
                        NordicAccessPerm.NRF_PERM_X in access_perm, permission.src
                    ),
                    secure=uicr.Value(
                        NordicAccessPerm.NRF_PERM_S in access_perm, permission.src
                    ),
                    non_secure_callable=uicr.Value(
                        NordicAccessPerm.NRF_PERM_NSC in access_perm, permission.src
                    ),
                )
                accesses.append(access)

        return accesses

    @DtMatch(
        compat="nordic,nrf-gpiote",
        status=STATUS_OWNED,
        regs=address_is_global_periph,
    )
    def visit_nrf_gpiote(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        addr = dt_reg_addr(node)
        channels = dt_channels_get(node)

        for num, secure in channels:
            self._builder.add_gpiote_channel(addr, uicr.Feature(num=num, secure=secure))

    @DtMatch(
        compat="nordic,nrf-dppic-global",
        status=STATUS_OWNED,
        regs=address_is_global_periph,
    )
    def visit_nrf_dppic(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        addr = dt_reg_addr(node)
        channels = dt_channels_get(node)
        channel_groups = dt_channels_get(
            node,
            owned_name="owned-channel-groups",
            nonsecure_name="nonsecure-channel-groups",
        )

        source_channels = dt_array_prop(node, "source-channels", [])
        sink_channels = dt_array_prop(node, "sink-channels", [])

        for num, secure in channels:
            self._builder.add_dppi_channel(addr, uicr.Feature(num=num, secure=secure))

        for num, secure in channel_groups:
            self._builder.add_dppi_channel_group(
                addr, uicr.Feature(num=num, secure=secure)
            )

        for num in source_channels:
            self._builder.add_dppi_link(
                addr,
                uicr.DppiLink(
                    num, uicr.Value(uicr.DppiLink.Direction.SOURCE, src=num.src)
                ),
            )

        for num in sink_channels:
            self._builder.add_dppi_link(
                addr,
                uicr.DppiLink(
                    num, uicr.Value(uicr.DppiLink.Direction.SINK, src=num.src)
                ),
            )

    @DtMatch(
        compat="nordic,nrf-ipct-global",
        status=STATUS_OWNED,
        regs=address_is_global_periph,
    )
    def visit_nrf_ipct_global(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        addr = dt_reg_addr(node)
        channels = dt_channels_get(node)

        for num, secure in channels:
            self._builder.add_ipct_channel(addr, uicr.Feature(num=num, secure=secure))

    @DtMatch(
        compat=re.compile("^nordic,nrf-ipct-((local)|(global))$"),
        status=STATUS_OWNED,
        props=["source-channel-links", "sink-channel-links"],
    )
    def visit_nrf_ipct(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        source_channel_links = dt_array_prop(node, "source-channel-links", [])
        if len(source_channel_links) % 3 != 0:
            raise BadDevicetreeError()
        sink_channel_links = dt_array_prop(node, "sink-channel-links", [])

        domain = dt_prop(node, "global-domain-id", None)
        if domain.val is None:
            addr = dt_reg_addr(node)
            domain_id = DomainID.from_address(addr.val)
            if domain_id is None:
                raise BadDevicetreeError(
                    f"Failed to determine domain ID for address 0x{addr.val:08x} "
                    f"specified on node {node.path}"
                )
            domain = uicr.Value(domain_id.value, src=addr.src)

        for source_ch, sink_domain, sink_ch in batched(source_channel_links, 3):
            link = uicr.IpcLink(
                source_domain=domain,
                source_channel_num=source_ch,
                sink_domain=sink_domain,
                sink_channel_num=sink_ch,
            )
            self._builder.add_ipc_link(link)

        for sink_ch, source_domain, source_ch in batched(sink_channel_links, 3):
            link = uicr.IpcLink(
                source_domain=source_domain,
                source_channel_num=source_ch,
                sink_domain=domain,
                sink_channel_num=sink_ch,
            )
            self._builder.add_ipc_link(link)

    @DtMatch(compat="zephyr,ipc-icmsg", status=STATUS_OWNED)
    def visit_ipc_icmsg(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        tx_idx = node.props["mbox-names"].val.index("tx")
        tx_mbox = node.props["mboxes"].val[tx_idx]

        tx_processor = dt_node_processor_id(tx_mbox.controller, interrupts_only=False)
        if (
            tx_processor.val != ProcessorID.SECURE
            or self._processor == ProcessorID.SYSCTRL
        ):
            # We only care about secure domain IPC
            return

        tx_region = node.props["tx-region"].val
        tx_addr = dt_reg_addr(tx_region)
        tx_size = dt_reg_size(tx_region)

        rx_region = node.props["rx-region"].val
        rx_addr = dt_reg_addr(rx_region)
        rx_size = dt_reg_size(rx_region)

        mailbox = uicr.Mailbox(
            tx_address=tx_addr, tx_size=tx_size, rx_address=rx_addr, rx_size=rx_size
        )
        self._builder.set_secure_mailbox(mailbox)

    @DtMatch(
        compat=re.compile(PROG_OWNER_COMPAT_PATTERN),
        status=STATUS_OWNED,
        regs=address_is_global_periph,
    )
    def visit_nordic_split_ownership_peripheral(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        addr = dt_reg_addr(node)
        processor = dt_node_processor_id(node, default=self._processor.value)
        secure = dt_node_is_secure(node)

        peripheral = uicr.Peripheral(
            address=addr, irq_processor=processor, secure=secure, dma_secure=secure
        )
        self._builder.add_peripheral(peripheral)

    @DtMatch(
        compat="nordic,nrf-grtc",
        status=STATUS_OWNED,
        regs=address_is_global_periph,
    )
    def visit_nrf_grtc(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        channels = dt_channels_get(node)
        for num, secure in channels:
            self._builder.add_grtc_channel(uicr.Feature(num=num, secure=secure))

    @DtMatch(status=STATUS_OWNED, props=[re.compile(r"^(.+-)?gpios$")])
    def visit_peripheral_with_gpios(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        # TODO: this could be solved better with e.g. a property matcher
        for name in node.props:
            if not re.fullmatch(r"^(.+-)?gpios$", name):
                continue

            for entry in dt_array_prop(node, name):
                gpio_node = entry.val.controller
                if "nordic,nrf-gpio" not in gpio_node.compats:
                    continue

                port = gpio_node.props["port"].val
                num = uicr.Value(entry.val.data["pin"], src=entry.src)
                secure = dt_node_is_secure(gpio_node)
                ctrlsel = dt_lookup_ctrlsel(
                    self._product, node.props[name], (port, num.val)
                )
                gpio_addr = dt_reg_addr(entry.val.controller)
                pin = uicr.Pin(num=num, secure=secure, ctrlsel=ctrlsel)
                self._builder.add_gpio_pin(gpio_addr, pin)

    @DtMatch(status=STATUS_OWNED, props=["pinctrl-0"])
    def visit_peripheral_with_pinctrls(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        secure = dt_node_is_secure(node)
        for pinctrl in node.pinctrls:
            for config_node in pinctrl.conf_nodes:
                for group_node in config_node.children.values():
                    for i, psel_val in enumerate(dt_array_prop(group_node, "psels")):
                        psel = NrfPsel.from_raw(psel_val.val)
                        gpio_node = find_gpio_node_by_port(
                            node.edt,
                            psel.port,
                            err_suffix=f" (referenced by {group_node.path}:psels[{i}])",
                        )
                        num = uicr.Value(psel.pin, src=psel_val.src)
                        ctrlsel = dt_lookup_ctrlsel(self._product, pinctrl, psel)
                        gpio_addr = dt_reg_addr(gpio_node)
                        pin = uicr.Pin(num=num, secure=secure, ctrlsel=ctrlsel)
                        self._builder.add_gpio_pin(gpio_addr, pin)

    @DtMatch(compat="nordic,nrf-saadc", status=STATUS_OWNED)
    def visit_nrf_saadc(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        secure = dt_node_is_secure(node)
        for name, child in node.children.items():
            if not name.startswith("channel"):
                continue
            for port, num in dt_lookup_adc_channel_pins(child):
                gpio_node = find_gpio_node_by_port(node.edt, port.val)
                gpio_addr = dt_reg_addr(gpio_node)
                pin = uicr.Pin(num, secure, ctrlsel=None)
                self._builder.add_gpio_pin(gpio_addr, pin)

    @DtMatch(compat=re.compile(r"^nordic,(lp)?comp$"), status=STATUS_OWNED)
    def visit_nrf_comp_lpcomp(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        secure = dt_node_is_secure(node)
        for port, num in dt_lookup_comp_lpcomp_pins(node):
            gpio_node = find_gpio_node_by_port(node.edt, port.val)
            gpio_addr = dt_reg_addr(gpio_node)
            pin = uicr.Pin(num, secure, ctrlsel=None)
            self._builder.add_gpio_pin(gpio_addr, pin)

    @DtMatch(compat="nordic,nrf-tddconf", status=STATUS_OWNED)
    def visit_nrf_tddconf(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        self_processor = self._processor.value

        for sources_name, sink_enum in [
            ("etbsources", uicr.Trace.Sink.ETB),
            ("tpiusources", uicr.Trace.Sink.TPIU),
            ("etrsources", uicr.Trace.Sink.ETR),
        ]:
            if sources_name not in node.props:
                continue

            sources_val = dt_prop(node, sources_name)
            tdd_sources = NordicTddSource(sources_val.val)
            for sources_bit, processor, source_enum in [
                (
                    NordicTddSource.SOURCE_STMMAINCORE,
                    self_processor,
                    uicr.Trace.Source.STM,
                ),
                (
                    NordicTddSource.SOURCE_ETMMAINCORE,
                    self_processor,
                    uicr.Trace.Source.ETM,
                ),
                (
                    NordicTddSource.SOURCE_STMPPR,
                    ProcessorID.PPR.value,
                    uicr.Trace.Source.STM,
                ),
                (
                    NordicTddSource.SOURCE_STMFLPR,
                    ProcessorID.FLPR.value,
                    uicr.Trace.Source.STM,
                ),
                (
                    NordicTddSource.SOURCE_STMHWEVENTS,
                    self_processor,
                    uicr.Trace.Source.STM_HW_EVENTS,
                ),
            ]:
                if sources_bit in tdd_sources:
                    trace = uicr.Trace(
                        processor=processor,
                        source=uicr.Value(source_enum, src=sources_val.src),
                        sink=uicr.Value(sink_enum, sources_val.src),
                    )
                    self._builder.add_trace_link(trace)

        if "portconfig" in node.props:
            speed = dt_prop(node, "portconfig")
            self._builder.set_trace_port_config(uicr.TracePortConfig(speed=speed))

        if "etrbuffer" in node.props:
            buffer_node = node.props["etrbuffer"].val
            address = dt_reg_addr(buffer_node)
            size = dt_reg_size(buffer_node)
            self._builder.set_etr_buffer(uicr.EtrBuffer(address=address, size=size))

    @DtMatch(compat="nordic,nrf-uicr-v2", status=STATUS_OWNED)
    def visit_nrf_uicr_v2(self, node: edtlib.Node, /) -> None:
        log.trace_function()

        node_domain = DomainID(node.props["domain"].val)
        if node_domain == self._domain:
            self.uicr_address = dt_reg_addr(node).val


def find_gpio_node_by_port(
    dt: edtlib.EDT, port: int, err_suffix: str = ""
) -> edtlib.Node:
    for gpio_node in dt.compat2nodes["nordic,nrf-gpio"]:
        if gpio_node.props["port"].val == port:
            return gpio_node
    raise BadDevicetreeError(
        f"Failed to find Nordic GPIO node with port {port}{err_suffix}"
    )


# Equivalent to itertools.batched(), using the recipe provided in the docs.
def batched(iterable: Iterable, n: int, *, strict: bool = False) -> Iterator:
    if n < 1:
        raise ValueError("n must be at least one")
    iterator = iter(iterable)
    while batch := tuple(islice(iterator, n)):
        if strict and len(batch) != n:
            raise ValueError("batched(): incomplete batch")
        yield batch


@overload
def dt_to_unit_src(dt_src: None, /) -> None: ...


@overload
def dt_to_unit_src(dt_src: DtLocation, /) -> uicr.UnitSrc: ...


def dt_to_unit_src(dt_src, /):
    """Create a unique string identifier for a devicetree location.

    Uses ';' as a separator as that is not allowed as part of a node/property name.

    Each unique location is converted to one of three forms:
    - A node path, e.g. "/my/path"
    - A node property path, e.g. "/my/path;my-prop"
    - A node property element path, e.g., "/my/path;my-prop;0"

    Lists of locations are converted into lists of identifiers.
    """
    match dt_src:
        case None:
            return None
        case edtlib.Node() as node:
            return f"{node.path}"
        case edtlib.Property() as prop:
            return f"{prop.node.path};{prop.name}"
        case (edtlib.Property() as prop, idx):
            return f"{prop.node.path};{prop.name};{idx}"
        case list():
            return [dt_to_unit_src(e) for e in dt_src]
        case _:
            raise ValueError(f"Unrecognized dt src {dt_src}")


class NoDefault: ...


NO_DEFAULT = NoDefault


def dt_prop(node: edtlib.Node, name: str, default: Any = NO_DEFAULT) -> uicr.Value[Any]:
    """Get the property value and identfier of a property.
    Optionally returns a default value.
    """
    try:
        prop = node.props[name]
    except KeyError:
        if default != NO_DEFAULT:
            return uicr.Value(default, src=[])
        raise

    return uicr.Value(prop.val, src=dt_to_unit_src(prop))


def dt_array_prop(
    node: edtlib.Node, name: str, default: list[Any] | Type[NO_DEFAULT] = NO_DEFAULT
) -> list[uicr.Value[Any]]:
    """Get the member values and identifiers of an array property.
    Optionally returns a default value.
    """
    try:
        prop = node.props[name]
    except KeyError:
        if default != NO_DEFAULT:
            return [uicr.Value(e, src=[]) for e in default]
        raise

    return [
        uicr.Value(v, src=dt_to_unit_src((prop, i))) for i, v in enumerate(prop.val)
    ]


def dt_partition_address_get(node: edtlib.Node) -> uicr.Value[int]:
    """Get the absolute address and identifiers for a "fixed-partitions" child node."""
    gparent = node.parent.parent
    gparent_addr = gparent.regs[0].addr
    self_addr = node.regs[0].addr
    self_src = dt_to_unit_src((node.props["reg"], 0))
    full_addr = gparent_addr + self_addr
    return uicr.Value(full_addr, src=self_src)


def dt_mem_node_is_secure(node: edtlib.Node) -> uicr.Value[bool]:
    """Get the secure permission and identifers for a memory node."""
    n = node
    while n is not None:
        if "nordic,owned-partitions" in n.compats or "nordic,owned-memory" in n.compats:
            if "nordic,access" in n.props:
                prop = (n.props["nordic,access"], 1)
                secure = bool(
                    n.props["nordic,access"].val[1] & NordicAccessPerm.NRF_PERM_S
                )
            else:
                prop = n.props["perm-secure"]
                secure = prop.val
            return uicr.Value(secure, src=dt_to_unit_src(prop))
        n = n.parent

    return uicr.Value(False, src=[])


def dt_reg_addr(node: edtlib.Node, index: int = 0) -> uicr.Value[int]:
    """Get a register address and property identifier for a node."""
    addr = node.regs[index].addr
    src = dt_to_unit_src((node.props["reg"], index))
    return uicr.Value(addr, src=src)


def dt_reg_size(node: edtlib.Node, index: int = 0) -> uicr.Value[int]:
    """Get a register size and property identifier for a node."""
    size = node.regs[index].size
    src = dt_to_unit_src((node.props["reg"], index))
    return uicr.Value(size, src=src)


def dt_node_processor_id(
    node: edtlib.Node,
    default: int | uicr.Value[int] | Type[NO_DEFAULT] = NO_DEFAULT,
    *,
    interrupts_only: bool = True,
) -> uicr.Value[int]:
    """Determine the processor ID a node corresponds to.
    This can be used to determine the CPU that will receive IRQs for a given peripheral.
    """
    processors: list[uicr.Value[int]] = []

    for irq in node.interrupts:
        irq_ctrl = irq.controller
        try:
            irq_processors = processor_ids_from_nodelabels(irq_ctrl.labels)
        except ValueError:
            continue
        for processor_id in irq_processors:
            processors.append(
                uicr.Value(processor_id.value, src=dt_to_unit_src(irq_ctrl))
            )

    if not processors:
        try:
            node_processors = processor_ids_from_nodelabels(node.labels)
            for processor_id in node_processors:
                processors.append(
                    uicr.Value(processor_id.value, src=dt_to_unit_src(node))
                )
        except ValueError:
            pass

    if not processors:
        if default == NO_DEFAULT:
            raise RuntimeError(
                "No unique processor ID could be found based on interrupt controllers "
                f"{'or nodelabels ' if not interrupts_only else ''}for node {node.path}"
            )
        if isinstance(default, uicr.Value):
            return default
        return uicr.Value(default, src=[])

    unique_processor_ids = set((i.val for i in processors))
    if len(unique_processor_ids) > 1:
        raise RuntimeError(
            f"Peripheral node {node.path} corresponds to multiple processors "
            f"({unique_processor_ids}), which is not supported."
        )

    return processors[0]


def processor_ids_from_nodelabels(labels: list[str]) -> list[ProcessorID]:
    """Deduce a processor ID from a list of devicetree nodelabels."""
    substring_processor = {
        "cpusec": ProcessorID.SECURE,
        "cpuapp": ProcessorID.APPLICATION,
        "cpurad": ProcessorID.RADIOCORE,
        "cpucell": ProcessorID.CELLCORE,
        "cpubbpr": ProcessorID.BBPR,
        "cpusys": ProcessorID.SYSCTRL,
        "cpuppr": ProcessorID.PPR,
        "cpuflpr": ProcessorID.FLPR,
    }
    processors = {
        processor_id
        for substring, processor_id in substring_processor.items()
        if any(substring in label for label in labels)
    }
    return list(processors)


def dt_channels_get(
    node: edtlib.Node,
    owned_name: str = "owned-channels",
    nonsecure_name: str = "nonsecure-channels",
) -> list[tuple[uicr.Value[int], uicr.Value[bool]]]:
    owned = []
    owned.extend(dt_array_prop(node, owned_name, default=[]))
    owned.extend(dt_array_prop(node, f"child-{owned_name}", default=[]))

    sec_lookup = {}
    if nonsecure_name in node.props:
        nonsecure = dt_array_prop(node, nonsecure_name)
        sec_lookup.update({v.val: uicr.Value(False, src=v.src) for v in nonsecure})

    default_sec = dt_node_is_secure(node)
    channels = []
    for ch in owned:
        sec = sec_lookup.setdefault(ch.val, default_sec)
        channels.append((ch, sec))

    return channels


def dt_node_is_secure(node: edtlib.Node) -> uicr.Value[bool]:
    if node.bus_node is not None and node.bus_node.regs:
        addr = dt_reg_addr(node.bus_node)
    elif node.regs:
        addr = dt_reg_addr(node)
    else:
        raise ValueError(
            f"Failed to determine security of {node.path} "
            "from the address of its bus node or itself"
        )

    secure = addr.val == secure_address_get(addr.val)

    return uicr.Value(secure, src=addr.src)


def dt_processor_id(devicetree: edtlib.EDT) -> ProcessorID:
    """Get processor information from a domain's devicetree."""
    cpus = [
        node
        for node in devicetree.get_node("/cpus").children.values()
        if node.name.startswith("cpu@")
    ]
    if len(cpus) != 1:
        raise RuntimeError(
            f"Expected exactly 1 'cpu' node, but devicetree contained {len(cpus)} nodes"
        )

    try:
        return ProcessorID(cpus[0].regs[0].addr)
    except Exception:
        raise RuntimeError(
            f"Devicetree 'cpu' node has invalid Processor ID {cpus[0].regs[0].addr}"
        )


@dataclass(frozen=True)
class NrfPsel:
    """Decoded NRF_PSEL values."""

    fun: int
    port: int
    pin: int

    @classmethod
    def from_raw(cls, psel_value: int) -> NrfPsel:
        """Decode a raw NRF_PSEL encoded int value to its individual parts."""
        port, pin = divmod(psel_value & (~NRF_PSEL_FUN_MASK), NRF_PSEL_GPIO_PIN_COUNT)
        fun = (psel_value & NRF_PSEL_FUN_MASK) >> NRF_PSEL_FUN_POS
        return NrfPsel(fun=fun, port=port, pin=pin)


@enum.unique
class NordicAccessPerm(enum.IntFlag):
    """Permission flags used with the "nordic,owned-memory" devicetree binding."""

    NRF_PERM_R = 1 << 0
    NRF_PERM_W = 1 << 1
    NRF_PERM_X = 1 << 2
    NRF_PERM_S = 1 << 3
    NRF_PERM_NSC = 1 << 4


@enum.unique
class NordicTddSource(enum.IntFlag):
    """Trace source flags used with the "nordic,tddconf" binding"""

    SOURCE_STMMAINCORE = 1 << 0
    SOURCE_ETMMAINCORE = 1 << 1
    SOURCE_STMHWEVENTS = 1 << 2
    SOURCE_STMPPR = 1 << 3
    SOURCE_STMFLPR = 1 << 4


# # Bit position of the function bits in the pinctrl pin value encoded from NRF_PSEL()
NRF_PSEL_FUN_POS = 24
# # Mask for the function bits in the pinctrl pin value encoded from NRF_PSEL()
NRF_PSEL_FUN_MASK = 0xFF << NRF_PSEL_FUN_POS
# Number of pins per port used in NRF_PSEL()
NRF_PSEL_GPIO_PIN_COUNT = 32

# Pin functions used with pinctrl, see include/zephyr/dt-bindings/pinctrl/nrf-pinctrl.h
# Only the functions relevant for CTRLSEL deduction have been included.
NRF_FUN_UART_TX = 0
NRF_FUN_UART_RX = 1
NRF_FUN_UART_RTS = 2
NRF_FUN_UART_CTS = 3
NRF_FUN_SPIM_SCK = 4
NRF_FUN_SPIM_MOSI = 5
NRF_FUN_SPIM_MISO = 6
NRF_FUN_SPIS_SCK = 7
NRF_FUN_SPIS_MOSI = 8
NRF_FUN_SPIS_MISO = 9
NRF_FUN_SPIS_CSN = 10
NRF_FUN_TWIM_SCL = 11
NRF_FUN_TWIM_SDA = 12
NRF_FUN_PWM_OUT0 = 22
NRF_FUN_PWM_OUT1 = 23
NRF_FUN_PWM_OUT2 = 24
NRF_FUN_PWM_OUT3 = 25
NRF_FUN_EXMIF_CK = 35
NRF_FUN_EXMIF_DQ0 = 36
NRF_FUN_EXMIF_DQ1 = 37
NRF_FUN_EXMIF_DQ2 = 38
NRF_FUN_EXMIF_DQ3 = 39
NRF_FUN_EXMIF_DQ4 = 40
NRF_FUN_EXMIF_DQ5 = 41
NRF_FUN_EXMIF_DQ6 = 42
NRF_FUN_EXMIF_DQ7 = 43
NRF_FUN_EXMIF_CS0 = 44
NRF_FUN_EXMIF_CS1 = 45
NRF_FUN_CAN_TX = 46
NRF_FUN_CAN_RX = 47
NRF_FUN_TWIS_SCL = 48
NRF_FUN_TWIS_SDA = 49
NRF_FUN_EXMIF_RWDS = 50
NRF_FUN_GRTC_CLKOUT_FAST = 55
NRF_FUN_GRTC_CLKOUT_32K = 56

# Under PR here https://github.com/nrfconnect/sdk-zephyr/pull/2314/files
NRF_FUN_TDM_SCK_M = 71
NRF_FUN_TDM_SCK_S = 72
NRF_FUN_TDM_FSYNC_M = 73
NRF_FUN_TDM_FSYNC_S = 74
NRF_FUN_TDM_SDIN = 75
NRF_FUN_TDM_SDOUT = 76
NRF_FUN_TDM_MCK = 77

# Deliberately defined as placeholders
NRF_FUN_I3C_SDA = 0xFF - 1
NRF_FUN_I3C_SCL = 0xFF


@dataclass(frozen=True)
class GpiosProp:
    """CTRLSEL lookup table entry for *-gpios properties"""

    name: str
    port: int
    pin: int


PINCTRL_CTRLSEL_LOOKUP_NRF54H20 = {
    # I3C120
    0x5F8D_3000: {
        # P2
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=0): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=6, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=6, pin=1): Ctrlsel.CAN_PWM_I3C,
    },
    # CAN120
    0x5F8D_8000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=9): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=8): Ctrlsel.CAN_PWM_I3C,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=5): Ctrlsel.CAN,
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=4): Ctrlsel.CAN,
    },
    # I3C121
    0x5F8D_E000: {
        # P2
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=2): Ctrlsel.CAN_PWM_I3C,
        # P7
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=7, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=7, pin=2): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM120
    0x5F8E_4000: {
        # P2
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=2, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=2, pin=5): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=2, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=2, pin=7): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=7): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=9): Ctrlsel.CAN_PWM_I3C,
        # P7
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=7, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=7, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=7, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=7, pin=7): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM130
    0x5F9A_4000: {
        # P9
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=9, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=9, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=9, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=9, pin=5): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIM130/SPIS130/TWIM130/TWIS130/UARTE130
    0x5F9A_5000: {
        # SPIM mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=4): Ctrlsel.SERIAL0,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=4): Ctrlsel.SERIAL0,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=4): Ctrlsel.SERIAL0,
        # TWIS mappings
        NrfPsel(fun=NRF_FUN_TWIS_SDA, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_TWIS_SCL, port=9, pin=4): Ctrlsel.SERIAL0,
        # UARTÈ mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=3): Ctrlsel.SERIAL0,
    },
    # SPIM131/SPIS131/TWIM131/TWIS131/UARTE131
    0x5F9A_6000: {
        # SPIM mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        GpiosProp(name="cs-gpios", port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # TWIS mappings
        NrfPsel(fun=NRF_FUN_TWIS_SDA, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TWIS_SCL, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # UARTÈ mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # SPIS120
    0x5F8E_5000: {
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=6, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=6, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=6, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=6, pin=0): Ctrlsel.SERIAL0,
    },
    # SPIM120/UARTE120
    0x5F8E_6000: {
        # SPIM P6 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=7): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=1): Ctrlsel.SERIAL0,
        # SPIM P7 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=7, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=7, pin=6): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=7, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=7, pin=3): Ctrlsel.SERIAL0,
        # SPIM P2 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=5): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=2, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=3): Ctrlsel.SERIAL0,
        # UARTÈ P6 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=6, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=6, pin=5): Ctrlsel.SERIAL0,
        # UARTÈ P7 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=7, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=7, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=7, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=7, pin=5): Ctrlsel.SERIAL0,
        # UARTÈ P2 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=2, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=2, pin=7): Ctrlsel.SERIAL0,
    },
    # SPIM121
    0x5F8E_7000: {
        # SPIM P6 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=13): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=12): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=6, pin=10): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=2): Ctrlsel.SERIAL0,
        # SPIM P7 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=7, pin=1): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=7, pin=0): Ctrlsel.EXMIF_RADIO_SERIAL1,
        GpiosProp(name="cs-gpios", port=7, pin=4): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=7, pin=2): Ctrlsel.EXMIF_RADIO_SERIAL1,
        # SPIM P2 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=11): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=10): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=2, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=2): Ctrlsel.SERIAL0,
    },
    # EXMIF
    0x5F09_5000: {
        NrfPsel(fun=NRF_FUN_EXMIF_CK, port=6, pin=0): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_RWDS, port=6, pin=2): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_CS0, port=6, pin=3): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ7, port=6, pin=4): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ1, port=6, pin=5): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ6, port=6, pin=6): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ0, port=6, pin=7): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ5, port=6, pin=8): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ3, port=6, pin=9): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ2, port=6, pin=10): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ4, port=6, pin=11): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_CS1, port=6, pin=13): Ctrlsel.EXMIF_RADIO_SERIAL1,
    },
    # TDM130
    0x5F99_2000: {
        # TDM P1 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=1, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=1, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=1, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=1, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=1, pin=5): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=1, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=1, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        # TDM P2 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=2, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=2, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=2, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=2, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=2, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # TDM131
    0x5F99_7000: {
        # TDM P1 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=1, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=1, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=1, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=1, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=1, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=1, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=1, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        # TDM P2 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=2, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=2, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=2, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=2, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=2, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=2, pin=7): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=2, pin=7): Ctrlsel.CAN_TDM_SERIAL2,
    },
}

PINCTRL_CTRLSEL_LOOKUP_NRF9280 = {
    # I3C120
    0x5F8D_3000: {
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=9): Ctrlsel.CAN_PWM_I3C,
    },
    # CAN120
    0x5F8D_8000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=5): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # CAN121
    0x5F8D_B000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=8): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # I3C121
    0x5F8D_E000: {
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=10): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=11): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM120
    0x5F8E_4000: {
        # P2
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=2, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=2, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=2, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=2, pin=3): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=7): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=9): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIS120
    0x5F8E_5000: {
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=6, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=6, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=6, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=6, pin=0): Ctrlsel.SERIAL0,
    },
    # SPIM120/UARTE120
    0x5F8E_6000: {
        # SPIM P2 mappings
        GpiosProp(name="cs-gpios", port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=0): Ctrlsel.SERIAL0,
        # SPIM P6 mappings
        GpiosProp(name="cs-gpios", port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=1): Ctrlsel.SERIAL0,
        # UARTE P2 mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=2, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=2, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=2, pin=4): Ctrlsel.SERIAL0,
        # UARTE P6 mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=6, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=6, pin=8): Ctrlsel.SERIAL0,
    },
    # SPIM121
    0x5F8E_7000: {
        # P2
        GpiosProp(name="cs-gpios", port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=1): Ctrlsel.SERIAL0,
        # P6
        GpiosProp(name="cs-gpios", port=6, pin=10): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=12): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=13): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=2): Ctrlsel.SERIAL0,
    },
    # PWM130
    0x5F9A_4000: {
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=9, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=9, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=9, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=9, pin=5): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIM130/SPIS130/TWIM130/TWIS130/UARTE130
    0x5F9A_5000: {
        # SPIM mappings
        GpiosProp(name="cs-gpios", port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=0): Ctrlsel.SERIAL0,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=0): Ctrlsel.SERIAL0,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=0): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=3): Ctrlsel.SERIAL0,
        # UARTE mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=3): Ctrlsel.SERIAL0,
    },
}

PINCTRL_CTRLSEL_LOOKUP = {
    Product.NRF54H20: PINCTRL_CTRLSEL_LOOKUP_NRF54H20,
    Product.NRF9280: PINCTRL_CTRLSEL_LOOKUP_NRF9280,
}


def dt_lookup_ctrlsel(
    product: Product,
    src: edtlib.PinCtrl | edtlib.Property,
    psel: NrfPsel | tuple[int, int],
) -> uicr.Value[int]:
    lut = PINCTRL_CTRLSEL_LOOKUP[product]

    if isinstance(src, edtlib.PinCtrl):
        identifier = dt_reg_addr(src.node)
        sub_entry = psel
    elif isinstance(src, edtlib.Property):
        assert isinstance(psel, tuple)
        try:
            identifier = dt_reg_addr(src.node)
        except IndexError:
            identifier = uicr.Value(src.node.label, src=dt_to_unit_src(src.node))
        sub_entry = GpiosProp(name=src.name, port=psel[0], pin=psel[1])
    else:
        raise ValueError(f"Unsupported GPIO pin source: {src}")

    if identifier.val in lut:
        ctrlsel = lut[identifier.val].get(sub_entry, CTRLSEL_DEFAULT)
    else:
        ctrlsel = CTRLSEL_DEFAULT

    log.debug(
        f"identifier={hex(identifier.val) if isinstance(identifier.val, int) else identifier.val}, "
        f"{sub_entry=} -> {ctrlsel=}"
    )

    return uicr.Value(ctrlsel, src=identifier.src)


NRF_SAADC_AIN0 = 1
NRF_SAADC_AIN1 = 2
NRF_SAADC_AIN2 = 3
NRF_SAADC_AIN3 = 4
NRF_SAADC_AIN4 = 5
NRF_SAADC_AIN5 = 6
NRF_SAADC_AIN6 = 7
NRF_SAADC_AIN7 = 8
NRF_SAADC_AIN8 = 9
NRF_SAADC_AIN9 = 10
NRF_SAADC_AIN10 = 11
NRF_SAADC_AIN11 = 12
NRF_SAADC_AIN12 = 13
NRF_SAADC_AIN13 = 14

CHANNEL_LOOKUP = {
    # SAADC
    0x5F98_2000: {
        NRF_SAADC_AIN0: (1, 0),
        NRF_SAADC_AIN1: (1, 1),
        NRF_SAADC_AIN2: (1, 2),
        NRF_SAADC_AIN3: (1, 3),
        NRF_SAADC_AIN4: (1, 4),
        NRF_SAADC_AIN5: (1, 5),
        NRF_SAADC_AIN6: (1, 6),
        NRF_SAADC_AIN7: (1, 7),
        NRF_SAADC_AIN8: (9, 0),
        NRF_SAADC_AIN9: (9, 1),
        NRF_SAADC_AIN10: (9, 2),
        NRF_SAADC_AIN11: (9, 3),
        NRF_SAADC_AIN12: (9, 4),
        NRF_SAADC_AIN13: (9, 5),
    }
}


def dt_lookup_adc_channel_pins(
    channel: edtlib.Node,
) -> list[tuple[uicr.Value[int], uicr.Value[int]]]:
    lut = CHANNEL_LOOKUP

    address = secure_address_get(channel.parent.regs[0].addr)
    if address not in lut:
        return []

    pins = []

    if "zephyr,input-positive" in channel.props:
        prop = channel.props["zephyr,input-positive"]
        if prop.val in lut[address]:
            port, pin = lut[address][prop.val]
            src = dt_to_unit_src(prop)
            log.debug(
                f"{address=:08x}, {channel.props['zephyr,input-positive']=} -> ({port=}, {pin=})"
            )
            pins.append((uicr.Value(port, src=src), uicr.Value(pin, src=src)))

    if "zephyr,input-negative" in channel.props:
        prop = channel.props["zephyr,input-negative"]
        if prop.val in lut[address]:
            port, pin = lut[address][prop.val]
            src = dt_to_unit_src(prop)
            log.debug(
                f"{address=:08x}, {channel.props['zephyr,input-negative']=} -> ({port=}, {pin=})"
            )
            pins.append((uicr.Value(port, src=src), uicr.Value(pin, src=src)))

    return pins


NRF_COMP_LPCOMP_AIN0 = "AIN0"
NRF_COMP_LPCOMP_AIN1 = "AIN1"
NRF_COMP_LPCOMP_AIN2 = "AIN2"
NRF_COMP_LPCOMP_AIN3 = "AIN3"
NRF_COMP_LPCOMP_AIN4 = "AIN4"
NRF_COMP_LPCOMP_AIN5 = "AIN5"
NRF_COMP_LPCOMP_AIN6 = "AIN6"
NRF_COMP_LPCOMP_AIN7 = "AIN7"
NRF_COMP_LPCOMP_AIN8 = "AIN8"
NRF_COMP_LPCOMP_AIN9 = "AIN9"

COMP_LPCOMP_LOOKUP = {
    # COMP/LPCOMP
    0x5F98_3000: {
        NRF_COMP_LPCOMP_AIN0: (1, 0),
        NRF_COMP_LPCOMP_AIN1: (1, 1),
        NRF_COMP_LPCOMP_AIN2: (1, 2),
        NRF_COMP_LPCOMP_AIN3: (1, 3),
        NRF_COMP_LPCOMP_AIN4: (1, 4),
        NRF_COMP_LPCOMP_AIN5: (1, 5),
        NRF_COMP_LPCOMP_AIN6: (1, 6),
        NRF_COMP_LPCOMP_AIN7: (1, 7),
        NRF_COMP_LPCOMP_AIN8: (9, 0),
        NRF_COMP_LPCOMP_AIN9: (9, 1),
    }
}


def dt_lookup_comp_lpcomp_pins(
    comp_lpcomp: edtlib.Node,
) -> list[tuple[uicr.Value[int], uicr.Value[int]]]:
    lut = COMP_LPCOMP_LOOKUP

    address = secure_address_get(comp_lpcomp.regs[0].addr)
    if address not in lut:
        return []

    pins = []

    if "psel" in comp_lpcomp.props:
        prop = comp_lpcomp.props["psel"]
        if prop.val in lut[address]:
            port, pin = lut[address][prop.val]
            src = dt_to_unit_src(prop)
            log.debug(f"{address=:08x}, psel -> ({port=}, {pin=})")
            pins.append((uicr.Value(port, src=src), uicr.Value(pin, src=src)))

    if "extrefsel" in comp_lpcomp.props:
        prop = comp_lpcomp.props["extrefsel"]
        if prop.val in lut[address]:
            port, pin = lut[address][prop.val]
            src = dt_to_unit_src(prop)
            log.debug(f"{address=:08x}, extrefsel -> ({port=}, {pin=})")
            pins.append((uicr.Value(port, src=src), uicr.Value(pin, src=src)))

    return pins



# Copyright (c) 2024 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import svd


def content_get_and_extend(field: svd.Field, msb_pos: int = 31) -> int:
    """Bit shift the field value so that it occupies the most significant bits of a 32-bit value"""
    return field.content << ((msb_pos + 1) - field.bit_width)


def content_truncate_and_set(field: svd.Field, content: int, msb_pos: int = 31) -> int:
    """Bit shift the field value so that it occupies the most significant bits of a 32-bit value"""
    truncate_bits = (msb_pos + 1) - field.bit_width
    truncate_mask = ~((1 << truncate_bits) - 1)

    if (content & truncate_mask) != content:
        raise ValueError(f"0x{content:08x} is not aligned to {(1 << truncate_bits)} B")

    new_content = content >> truncate_bits
    field.content = new_content

    return new_content


class SvdEnum:
    """Helper class for constructing commonly used SVD enums."""

    ENABLED = "Enabled"
    DISABLED = "Disabled"

    @classmethod
    def enabled_if(cls, is_enabled: bool) -> str:
        return cls.ENABLED if is_enabled else cls.DISABLED

    ALLOWED = "Allowed"
    NOT_ALLOWED = "NotAllowed"

    @classmethod
    def allowed_if(cls, is_allowed: bool) -> str:
        return cls.ALLOWED if is_allowed else cls.NOT_ALLOWED

    SECURE = "Secure"
    NONSECURE = "NonSecure"

    @classmethod
    def secure_if(cls, is_secure: bool) -> str:
        return cls.SECURE if is_secure else cls.NONSECURE

    REQUESTED = "Requested"
    NOT_REQUESTED = "NotRequested"

    @classmethod
    def requested_if(cls, is_requested: bool) -> str:
        return cls.REQUESTED if is_requested else cls.NOT_REQUESTED

    OWN = "Own"
    NOT_OWN = "NotOwn"

    SINK = "Sink"
    SOURCE = "Source"

    LINKED = "Linked"
    NOT_LINKED = "NotLinked"



# Copyright (c) 2022 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

"""Various types and definitions for the Haltium platform."""

from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import Optional


@enum.unique
class Product(enum.Enum):
    """Enumeration of supported product names in the Haltium family."""

    NRF54H20 = enum.auto()
    NRF9280 = enum.auto()

    @classmethod
    def _missing_(cls, value: object) -> Optional[Product]:
        """Custom enum behavior to support case insensitive values."""
        if not isinstance(value, str):
            return None
        return cls.__members__.get(value.upper())

    @property
    def product_code(self) -> ProductCode:
        match self:
            case Product.NRF54H20:
                return ProductCode(part_number=0x16, revision=0x2)
            case Product.NRF9280:
                return ProductCode(part_number=0x12, revision=0x1)
            case _:
                raise ValueError(f"No product code found for {self}")


@dataclass
class ProductCode:
    part_number: int
    revision: int


@enum.unique
class AddressOffset(enum.IntEnum):
    """Address bit offsets, defined by Address Format of the product specification."""

    REGION = 29
    SECURITY = 28
    DOMAINID = 24
    ADDR = 23


def secure_address_get(address: int) -> int:
    """Get the TrustZone secure address for the given address"""
    return address | (1 << AddressOffset.SECURITY)


@enum.unique
class AddressRegion(enum.IntEnum):
    """Address regions, defined by Address Format of the product specification."""

    PROGRAM = 0
    DATA = 1
    PERIPHERAL = 2
    EXT_XIP = 3
    EXT_XIP_ENCRYPTED = 4
    STM = 5
    CPU = 7

    @classmethod
    def from_address(cls, address: int) -> AddressRegion:
        """Get the address region of an address."""
        return cls((address >> AddressOffset.REGION) & 0b111)


@enum.unique
class DomainID(enum.IntEnum):
    """Domain IDs in Haltium products."""

    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    GLOBAL_FAST = 12
    GLOBAL_SLOW = 13
    GLOBAL_ = 14
    GLOBAL = 15

    @classmethod
    def from_address(cls, address: int) -> DomainID:
        """Get the domain ID of an address."""
        return cls((address >> AddressOffset.DOMAINID) & 0xF)

    @classmethod
    def from_processor(cls, processor: ProcessorID | int) -> DomainID:
        """Get the domain ID corresponding to a processor ID."""
        processor_domain = {
            ProcessorID.SECURE: cls.SECURE,
            ProcessorID.APPLICATION: cls.APPLICATION,
            ProcessorID.RADIOCORE: cls.RADIOCORE,
            ProcessorID.CELLCORE: cls.CELLCORE,
            ProcessorID.CELLDSP: cls.CELLDSP,
            ProcessorID.CELLRF: cls.CELLRF,
            ProcessorID.ISIMCORE: cls.ISIMCORE,
            ProcessorID.SYSCTRL: cls.GLOBAL_FAST,
            ProcessorID.PPR: cls.GLOBAL_SLOW,
            ProcessorID.FLPR: cls.GLOBAL_,
        }
        return processor_domain[ProcessorID(processor)]


@enum.unique
class OwnerID(enum.IntEnum):
    """Enumeration of ownership IDs in haltium products."""

    NONE = 0
    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELL = 4
    ISIMCORE = 5
    SYSCTRL = 8

    @classmethod
    def from_domain(cls, domain: DomainID | int) -> OwnerID:
        """Get the owner ID corresponding to a domain ID."""
        domain_owner = {
            DomainID.SECURE: cls.SECURE,
            DomainID.APPLICATION: cls.APPLICATION,
            DomainID.RADIOCORE: cls.RADIOCORE,
            DomainID.CELLCORE: cls.CELL,
            DomainID.CELLDSP: cls.CELL,
            DomainID.CELLRF: cls.CELL,
            DomainID.ISIMCORE: cls.ISIMCORE,
            DomainID.GLOBAL_FAST: cls.SYSCTRL,
        }
        return domain_owner[DomainID(domain)]

    @classmethod
    def from_processor(cls, processor: ProcessorID | int) -> OwnerID:
        """Get the owner ID corresponding to a processor ID."""
        return cls.from_domain(DomainID.from_processor(processor))


@enum.unique
class ProcessorID(enum.IntEnum):
    """Processor IDs in haltium products."""

    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    BBPR = 11
    SYSCTRL = 12
    PPR = 13
    FLPR = 14

    @classmethod
    def from_domain(cls, domain: DomainID | int) -> ProcessorID:
        """Get the processor ID corresponding to a domain ID."""
        domain_processor = {
            DomainID.SECURE: cls.SECURE,
            DomainID.APPLICATION: cls.APPLICATION,
            DomainID.RADIOCORE: cls.RADIOCORE,
            DomainID.CELLCORE: cls.CELLCORE,
            DomainID.CELLDSP: cls.CELLDSP,
            DomainID.CELLRF: cls.CELLRF,
            DomainID.ISIMCORE: cls.ISIMCORE,
            DomainID.GLOBAL_FAST: cls.SYSCTRL,
            DomainID.GLOBAL_SLOW: cls.PPR,
            DomainID.GLOBAL_: cls.FLPR,
        }
        return domain_processor[DomainID(domain)]


@enum.unique
class Ctrlsel(enum.IntEnum):
    """
    Enumeration of GPIO.PIN_CNF[n].CTRLSEL values.
    The list here may not be exhaustive.
    """

    GPIO = 0
    VPR_GRC = 1
    CAN_PWM_I3C = 2
    SERIAL0 = 3
    EXMIF_RADIO_SERIAL1 = 4
    CAN_TDM_SERIAL2 = 5
    CAN = 6
    TND = 7


# Default CTRLSEL value indicating that CTRLSEL should not be used
CTRLSEL_DEFAULT = Ctrlsel.GPIO



# Copyright (c) 2022 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import inspect
import logging
from itertools import pairwise
from typing import Iterable


def memory_to_bytes(memory_iter: Iterable[tuple[int, int]], padding: int) -> bytearray:
    out = bytearray()

    for (addr_a, value_a), (addr_b, value_b) in pairwise(memory_iter):
        if not out:
            out.append(value_a)

        num_empty = addr_b - addr_a - 1
        if num_empty > 0:
            out.extend([padding] * num_empty)

        out.append(value_b)

    return out


def ranges_overlap_inclusive(
    a_start: int, a_end: int, b_start: int, b_end: int
) -> bool:
    return a_end >= b_start and b_end >= a_start


class RegtoolLogger(logging.Logger):
    NOTSET = logging.NOTSET
    TRACE = 5
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def trace_function(self) -> None:
        self.log(RegtoolLogger.TRACE, "%s", _get_func_call_formatter(1))


def _get_func_call_formatter(stack_offset: int = 0) -> _FunctionCallFormatter:
    frame_info = inspect.stack()[stack_offset + 1]
    return _FunctionCallFormatter(frame_info)


class _FunctionCallFormatter:
    def __init__(self, frame_info: inspect.FrameInfo) -> None:
        self._frame_info = frame_info

    def __call__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        args = inspect.getargvalues(self._frame_info.frame)
        arg_strings = [f"{k}={v}" for k, v in args.locals.items() if k in args.args]
        return f"{self._frame_info.function}({', '.join(arg_strings)})"



# Copyright (c) 2022 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import typing

from ._common import RegtoolLogger

_PACKAGE_NAME = "nrf-regtool"


def _get_version() -> str:
    import importlib.metadata

    return importlib.metadata.version(_PACKAGE_NAME)


__version__ = _get_version()


def _init_logger() -> RegtoolLogger:
    formatter = logging.Formatter("{message}", style="{")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    # Override the class returned by getLogger() below
    orig_class = logging.getLoggerClass()
    logging.setLoggerClass(RegtoolLogger)

    logger = typing.cast(RegtoolLogger, logging.getLogger("nrfregtool"))
    logger.setLevel(logging.ERROR)
    logger.addHandler(handler)

    # Reset custom logger class
    logging.setLoggerClass(orig_class)

    return logger


# logging.Logger instance used for log output from nrf-regtool
log = _init_logger()


# Public exports
__all__ = [
    "RegtoolLogger",
    "log",
    "__version__",
]



# Copyright (c) 2024 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

"""APIs for programmatically generating a UICR structure."""

from __future__ import annotations

import importlib.resources
import json
import textwrap
from dataclasses import dataclass
from enum import Enum
from typing import NamedTuple, Optional, TypeVar

import svd

from nrfregtool import log
from nrfregtool.platform import (
    DomainID,
    OwnerID,
    Product,
    ProductCode,
    ProcessorID,
    secure_address_get,
    CTRLSEL_DEFAULT,
)
from nrfregtool._common import (
    memory_to_bytes,
    ranges_overlap_inclusive,
)
from nrfregtool._svd_util import (
    SvdEnum,
    content_get_and_extend,
)
from ._common import (
    RegSrcMap,
    UnitSrc,
    Value,
)

# This API is part of the public library API.
# Any change to this API should usually result in a change of the package version.
# In general, do not break this API unless there is a very good reason for it.
# When adding support for new UICR format versions, add it in a backward compatible way.
# The API should be fully backward compatible with every UICR format change.

# Magic word used to identify the UICR header structure
UICR_MAGIC = 0x55494352


# Version tuple type
class UicrVersion(NamedTuple):
    major: int
    minor: int
    patch: int
    tweak: int


# List of all supported UICR format versions
UICR_VERSIONS: list[UicrVersion] = [UicrVersion(major=1, minor=0, patch=0, tweak=0)]

# Tuple containing the minimum supported UICR format version
UICR_VERSION_MIN = UICR_VERSIONS[0]

# Tuple containing the maximum supported UICR format version
UICR_VERSION_MAX = UICR_VERSIONS[-1]

# Tuple containing the default UICR format version
UICR_VERSION_DEFAULT = UICR_VERSION_MAX

# Fill value for empty non-volatile memory
_EMPTY_NVM_SPACE = 0xFF


class UicrError(RuntimeError):
    """Error raised by the UICR builder API"""

    ...


class UicrVersionError(UicrError):
    """Error raised when the UICR builder API is given an unsupported UICR version"""

    def __init__(
        self,
        got_version: UicrVersion,
        /,
    ) -> None:
        supported_versions = ["'{}.{}.{}+{}'".format(*v) for v in UICR_VERSIONS]
        super(UicrVersionError, self).__init__(
            "Unsupported UICR version "
            + "'{}.{}.{}+{}'. ".format(*got_version)
            + f"This package supports versions {supported_versions}"
        )


T = TypeVar("T")


def _value_unpack(value: T | Value[T]) -> tuple[T, Optional[UnitSrc]]:
    if isinstance(value, Value):
        return value.val, value.src
    return value, None


@dataclass
class Header:
    """Metadata describing the UICR structure itself.

    :var part_code: Part code of the product the UICR is generated for.
    :var hardware_revision: Hardware revision of the product the UICR is generated for.
    :var domain: domain ID of the local domain the UICR is associated with.
    """

    part_code: int | Value[int]
    hardware_revision: int | Value[int]
    domain: int | Value[int]


@dataclass
class Memory:
    """Owned global memory region.

    :var address: start address of the region.
    :var size: size of the region.
    :var access: list of access permissions for the region.
    """

    address: int | Value[int]
    size: int | Value[int]
    access: list[Access]

    @dataclass
    class Access:
        """Memory region access permissions.

        :var owner: owner ID of the owner the permissions are given to.
        :var read: the owner has read permissions.
        :var write: the owner has write permissions.
        :var execute: the owner has execute permissions.
        :var secure: access is limited to secure mode.
        :var non_secure_callable: memory region is non-secure callable.
        """

        owner: int | Value[int]
        read: bool | Value[bool] = False
        write: bool | Value[bool] = False
        execute: bool | Value[bool] = False
        secure: bool | Value[bool] = False
        non_secure_callable: bool | Value[bool] = False


@dataclass
class Mailbox:
    """SSF IPC buffer configuration.

    :var tx_address: start address of the TX buffer.
    :var tx_size: size of the TX buffer.
    :var rx_address: start address of the RX buffer.
    :var rx_size: size of the RX buffer.
    """

    tx_address: int | Value[int]
    tx_size: int | Value[int]
    rx_address: int | Value[int]
    rx_size: int | Value[int]


@dataclass
class Peripheral:
    """Owned global domain peripheral configuration.

    :var address: base address of the peripheral.
    :var irq_processor: processor ID of the processor that will receive the peripheral IRQ.
    :var secure: access is limited to secure mode.
    :var dma_secure: DMA is limited to secure mode, if applicable.
    """

    address: int | Value[int]
    irq_processor: int | Value[int]
    secure: bool | Value[bool] = False
    dma_secure: bool | Value[bool] = False


@dataclass
class Pin:
    """Owned GPIO pin configuration.

    :var num: pin number.
    :var secure: access is limited to secure mode.
    :var ctrlsel: CTRLSEL setting for the pin.
    """

    num: int | Value[int]
    secure: bool | Value[bool] = False
    ctrlsel: int | Value[int] = CTRLSEL_DEFAULT


@dataclass
class Feature:
    """Generic configuration of an owned feature in a global domain peripheral with split ownership.

    :var num: feature index.
    :var secure: access is limited to secure mode.
    """

    num: int | Value[int]
    secure: bool | Value[bool] = False


@dataclass
class DppiLink:
    """A link between a PPI channel in one PPI domain with the neighboring PPI domain.

    :var channel_num: PPI channel number.
    :var direction: link direction.
    """

    channel_num: int | Value[int]
    direction: Direction | Value[Direction]

    class Direction(str, Enum):
        SOURCE = "source"
        SINK = "sink"


@dataclass
class IpcLink:
    """A link between two IPCT channels.

    :var source_domain: domain ID of the source IPCT peripheral.
    :var source_channel_num: channel number in the source IPCT peripheral.
    :var sink_domain: domain ID of the sink IPCT peripheral.
    :var sink_channel_num: channel number in the sink IPCT peripheral.
    """

    source_domain: int | Value[int]
    source_channel_num: int | Value[int]
    sink_domain: int | Value[int]
    sink_channel_num: int | Value[int]


@dataclass
class Trace:
    """A link between a TDD trace source and sink.

    :var processor: processor ID of the processor the trace source corresponds to.
    :var source: trace source.
    :var sink: trace sink.
    """

    processor: int | Value[int]
    source: Source | Value[Source]
    sink: Sink | Value[Sink]

    class Source(str, Enum):
        STM = "stm"
        ETM = "etm"
        STM_LOCAL = "stm-local"
        STM_HW_EVENTS = "stm-hw-events"

    class Sink(str, Enum):
        ETB = "etb"
        TPIU = "tpiu"
        ETR = "etr"


@dataclass
class TracePortConfig:
    """Trace port configuration.

    :var speed: TPIU clock divisor.
    """

    speed: int | Value[int]


@dataclass
class EtrBuffer:
    """ETR buffer configuration.

    :var address: start address of the ETR buffer.
    :var size: size of the ETR buffer.
    """

    address: int | Value[int]
    size: int | Value[int]


class Uicr:
    """Builder used to construct a UICR programmatically
    and output the structure in various formats.
    """

    def __init__(
        self, version: UicrVersion | Value[UicrVersion] = UICR_VERSION_DEFAULT
    ) -> None:
        log.trace_function()

        version, version_src = _value_unpack(version)
        if version not in UICR_VERSIONS:
            raise UicrVersionError(version)

        uicr_svd = importlib.resources.files("nrfregtool.resources").joinpath(
            "uicr.svd"
        )
        with importlib.resources.as_file(uicr_svd) as svd_file:
            device = svd.parse(
                svd_file,
                options=svd.Options(
                    parent_relative_cluster_address=True,
                ),
            )

        self._uicr = device["UICR"]

        self._map = RegSrcMap()

        header = self._uicr["HEADER"]

        self._map.reg_set(header["MAGIC"], UICR_MAGIC, None)

        self._map.reg_set(header["VERSION"]["MAJOR"], version.major, version_src)
        self._map.reg_set(header["VERSION"]["MINOR"], version.minor, version_src)
        self._map.reg_set(header["VERSION"]["PATCH"], version.patch, version_src)
        self._map.reg_set(header["VERSION"]["TWEAK"], version.tweak, version_src)

    def build_debug_info(self) -> str:
        """Create a string containing debug information used for improved error reporting.

        The returned string should be treated as an opaque object.

        :returns: debug information string.
        """
        return json.dumps(self._map.src_dict, indent=2)

    def build_bytes(self) -> bytearray:
        """Encode the UICR as bytes.

        :returns: byte representation of the UICR content.
        """
        log.trace_function()

        return memory_to_bytes(self._uicr.memory_iter(), _EMPTY_NVM_SPACE)

    @classmethod
    def from_bytes(cls, memory: bytearray) -> Uicr:
        """Initialize the UICR from bytes.

        An exception is raised if the length of the input exceeds the size of the UICR structure.
        """
        log.trace_function()

        builder = cls()
        uicr = builder._uicr

        req_address_bounds = uicr.address_bounds
        uicr_size = req_address_bounds[1] - req_address_bounds[0]

        input_size = len(memory)
        if input_size > uicr_size:
            raise UicrError(
                f"Memory with size 0x{input_size:x} "
                f"does not fit in the UICR structure with size 0x{uicr_size:x}"
            )

        for reg in uicr.register_iter(leaf_only=True):
            try:
                content_parts = [memory[a] for a in reg.address_range]
            except KeyError:
                if reg.path[0] == "HEADER":
                    raise UicrError("Memory does not contain a full UICR header")

                # TODO: can we fail for other regs as well here?
                continue

            content = int.from_bytes(content_parts, "little")

            # We don't currently support UICR version conversion.
            # Therefore we check that the blob is the same version that is used by the constructor.
            if reg.path == svd.EPath("HEADER.MAGIC"):
                if content != UICR_MAGIC:
                    raise UicrError(
                        "Memory magic value does not match the UICR magic value. "
                        f"Expected 0x{reg.content:08x}, got 0x{content:08x}"
                    )
            # Need to be able to set reset values
            reg.unconstrain()
            reg.content = content

            if reg.path == svd.EPath("HEADER.VERSION"):
                v_mem = (
                    reg["MAJOR"].content,
                    reg["MINOR"].content,
                    reg["PATCH"].content,
                    reg["TWEAK"].content,
                )
                if v_mem not in UICR_VERSIONS:
                    raise UicrVersionError(v_mem)

        return builder

    def set_header(self, header: Header) -> Uicr:
        """Configure the UICR metadata.

        :param header: header data.
        :returns: builder instance.
        """
        log.trace_function()

        reg = self._uicr["HEADER"]

        part_code, part_code_src = _value_unpack(header.part_code)
        self._map.reg_set(reg["PARTNO"]["PARTNO"], part_code, part_code_src)

        rev, rev_src = _value_unpack(header.hardware_revision)
        self._map.reg_set(reg["HWREVISION"]["HWREVISION"], rev, rev_src)

        domain, domain_src = _value_unpack(header.domain)
        self._map.reg_set(reg["DOMAIN"]["DOMAIN"], domain, domain_src)

        return self

    def set_secure_vtor(self, vtor_addr: int | Value[int]) -> Uicr:
        """Set the initial Vector Table Offset Register for the secure image.

        :param vtor_addr: VTOR address.
        :returns: builder instance.
        """
        log.trace_function()

        self._map.reg_set(
            self._uicr["INITSVTOR"]["INITSVTOR"], *_value_unpack(vtor_addr)
        )
        return self

    def set_nonsecure_vtor(self, vtor_addr: int | Value[int]) -> Uicr:
        """Set the initial Vector Table Offset Register for the nonsecure image.

        :param vtor_addr: VTOR address.
        :returns: builder instance.
        """
        log.trace_function()

        self._map.reg_set(
            self._uicr["INITNSVTOR"]["INITNSVTOR"], *_value_unpack(vtor_addr)
        )
        return self

    def add_memory(self, memory: Memory) -> Uicr:
        """Add an owned global memory region for the local domain.

        :param memory: memory region.
        :returns: builder instance.
        """
        log.trace_function()

        for access in memory.access:
            self._add_memory_with_access(memory, access)
        return self

    def _add_memory_with_access(
        self,
        memory: Memory,
        access: Memory.Access,
    ) -> None:
        addr, addr_src = _value_unpack(memory.address)
        size, size_src = _value_unpack(memory.size)
        addr_end_exc = addr + size
        read, read_src = _value_unpack(access.read)
        write, write_src = _value_unpack(access.write)
        execute, execute_src = _value_unpack(access.execute)
        secure, secure_src = _value_unpack(access.secure)
        owner, owner_src = _value_unpack(access.owner)
        nsc, nsc_src = _value_unpack(access.non_secure_callable)

        mem_entries = self._uicr["MEM"]
        mem_mergeable = []
        for i, mem in enumerate(mem_entries):
            mem_config0 = mem["CONFIG0"]
            mem_config1 = mem["CONFIG1"]

            if not mem_config0.modified:
                # No more occupied entries after this point
                mem_count = i
                # If we found overlapping entries, then we will overwrite the first one.
                # Otherwise, we will overwrite the first vacant entry.
                mem_target = mem_mergeable[0] if mem_mergeable else mem
                break

            mem_addr = content_get_and_extend(mem_config0["ADDRESS"])
            mem_size = content_get_and_extend(mem_config1["SIZE"])
            mem_addr_end_exc = mem_addr + mem_size

            # Check if we can merge this entry with the existing entry
            if ranges_overlap_inclusive(
                addr,
                addr_end_exc,
                mem_addr,
                mem_addr_end_exc,
            ) and (
                mem_config1["OWNERID"].content,
                mem_config0["READ"].modified,
                mem_config0["WRITE"].modified,
                mem_config0["EXECUTE"].modified,
                not mem_config0["SECURE"].modified,
                mem_config0["NSC"].modified,
            ) == (
                owner,
                read,
                write,
                execute,
                secure,
                nsc,
            ):
                mem_mergeable.append(mem)
        else:
            # All entries are occupied
            mem_count = len(mem_entries)
            if not mem_mergeable:
                raise UicrError(
                    f"Too many UICR memory configurations (max: {mem_count})"
                )
            mem_target = mem_mergeable[0]

        # Merge existing entries
        for mem in mem_mergeable:
            mem_addr = content_get_and_extend(mem["CONFIG0"]["ADDRESS"])
            mem_size = content_get_and_extend(mem["CONFIG1"]["SIZE"])
            mem_addr_end_exc = mem_addr + mem_size

            addr = min(addr, mem_addr)
            addr_end_exc = max(addr_end_exc, mem_addr_end_exc)
            size = addr_end_exc - addr

            if mem != mem_target:
                self._map.reg_move(mem_target, mem)

        # Write final values into the target entry
        mem_config0 = mem_target["CONFIG0"]
        mem_config1 = mem_target["CONFIG1"]
        self._map.reg_set(mem_config0["ADDRESS"], addr, addr_src, lsb_truncate=True)
        self._map.reg_set(
            mem_config0["READ"], SvdEnum.allowed_if(access.read), read_src
        )
        self._map.reg_set(mem_config0["WRITE"], SvdEnum.allowed_if(write), write_src)
        self._map.reg_set(
            mem_config0["EXECUTE"], SvdEnum.allowed_if(execute), execute_src
        )
        self._map.reg_set(mem_config0["SECURE"], SvdEnum.secure_if(secure), secure_src)
        self._map.reg_set(mem_config0["NSC"], SvdEnum.enabled_if(nsc), nsc_src)
        self._map.reg_set(mem_config1["SIZE"], size, size_src, lsb_truncate=True)
        self._map.reg_set(mem_config1["OWNERID"], owner, owner_src)

        # Move occupied entries to fill any gaps left by the merging procedure
        mem_gaps = mem_mergeable[1:]
        if not mem_gaps:
            return
        for mem in mem_entries[mem_count - 1 :: -1]:
            if mem == mem_gaps[-1]:
                # Nothing to fill
                mem_gaps.pop()
            else:
                self._map.reg_move(mem_gaps.pop(0), mem)

            if not mem_gaps:
                break

    def add_peripheral(self, peripheral: Peripheral) -> Uicr:
        """Add and configure an owned global domain peripheral for the local domain.

        :param peripheral: peripheral.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(peripheral.address)
        secure_addr = secure_address_get(addr)
        for reg in self._uicr["PERIPH"]:
            cfg = reg["CONFIG"]
            if (
                not cfg.modified
                or content_get_and_extend(cfg["ADDRESS"]) == secure_addr
            ):
                break
        else:
            raise UicrError(
                f"Too many entries in PERIPH (max: {len(self._uicr['PERIPH'])})"
            )

        if not reg["CONFIG"].modified:
            self._map.reg_set(
                reg["CONFIG"]["ADDRESS"], secure_addr, addr_src, lsb_truncate=True
            )

        irq_processor, irq_processor_src = _value_unpack(peripheral.irq_processor)
        reg_processor = reg["CONFIG"]["PROCESSOR"]
        if reg_processor.modified and reg_processor.content != irq_processor:
            raise UicrError(
                f"Conflicting IRQ processor for peripheral 0x{addr:08x}: "
                f"{reg_processor.content}, {irq_processor}"
            )
        self._map.reg_set(reg_processor, irq_processor, irq_processor_src)

        secure, secure_src = _value_unpack(peripheral.secure)
        self._map.reg_set(
            reg["CONFIG"]["SECURE"],
            _resolve_secure(reg["CONFIG"]["SECURE"], secure),
            secure_src,
        )

        dma_secure, dma_secure_src = _value_unpack(peripheral.dma_secure)
        self._map.reg_set(
            reg["CONFIG"]["DMASEC"],
            _resolve_secure(reg["CONFIG"]["DMASEC"], dma_secure),
            dma_secure_src,
        )

        return self

    def add_grtc_channel(self, channel: Feature) -> Uicr:
        """Add an owned GRTC channel for the local domain.

        :param channel: GRTC channel.
        :returns: builder instance.
        """
        log.trace_function()

        reg = self._uicr["GRTC"]
        num, num_src = _value_unpack(channel.num)
        self._map.reg_set(reg["CC"]["OWN"][f"CC_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(channel.secure)
        self._map.reg_set(
            reg["CC"]["SECURE"][f"CC_{num}"],
            _resolve_secure(reg["CC"]["SECURE"][f"CC_{num}"], secure),
            secure_src,
        )

        return self

    def add_gpiote_channel(
        self,
        gpiote_addr: int | Value[int],
        channel: Feature,
    ) -> Uicr:
        """Add an owned global domain GPIOTE channel for the local domain.

        :param gpiote_addr: base address of the GPIOTE peripheral.
        :param channel: GPIOTE channel.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(gpiote_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["GPIOTE"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(channel.num)
        self._map.reg_set(reg["CH"]["OWN"][f"CH_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(channel.secure)
        self._map.reg_set(
            reg["CH"]["SECURE"][f"CH_{num}"],
            _resolve_secure(reg["CH"]["SECURE"][f"CH_{num}"], secure),
            secure_src,
        )

        return self

    def add_dppi_channel(
        self,
        dppic_addr: int | Value[int],
        channel: Feature,
    ) -> Uicr:
        """Add an owned global domain DPPI channel for the local domain.

        :param dppic_addr: base address of the DPPIC peripheral managing the PPI domain.
        :param channel: PPI channel.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(dppic_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["DPPI"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(channel.num)
        self._map.reg_set(reg["CH"]["OWN"][f"CH_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(channel.secure)
        self._map.reg_set(
            reg["CH"]["SECURE"][f"CH_{num}"],
            _resolve_secure(reg["CH"]["SECURE"][f"CH_{num}"], secure),
            secure_src,
        )

        return self

    def add_dppi_channel_group(
        self,
        dppic_addr: int | Value[int],
        channel_group: Feature,
    ) -> Uicr:
        """Add an owned global domain DPPI group for the local domain.

        :param dppic_addr: base address of the DPPIC peripheral managing the PPI domain.
        :param channel_group: PPI channel group.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(dppic_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["DPPI"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(channel_group.num)
        self._map.reg_set(reg["CHG"]["OWN"][f"CHG_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(channel_group.secure)
        self._map.reg_set(
            reg["CHG"]["SECURE"][f"CHG_{num}"],
            _resolve_secure(reg["CHG"]["SECURE"][f"CHG_{num}"], secure),
            secure_src,
        )

        return self

    def add_dppi_link(
        self,
        dppic_addr: int | Value[int],
        link: DppiLink,
    ) -> Uicr:
        """Link a global domain DPPI channel with a neighboring PPI domain.

        :param dppic_addr: base address of the DPPIC peripheral managing the PPI domain.
        :param link: PPI channel link description.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(dppic_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["DPPI"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        direction, direction_src = _value_unpack(link.direction)
        if direction == DppiLink.Direction.SINK:
            dir_enum = SvdEnum.SINK
        elif direction == DppiLink.Direction.SOURCE:
            dir_enum = SvdEnum.SOURCE
        else:
            raise ValueError(f"Unrecognized DPPI link direction: {direction}")

        num, num_src = _value_unpack(link.channel_num)
        reg_en = reg["CH"]["LINK"]["EN"][f"CH_{num}"]
        reg_dir = reg["CH"]["LINK"]["DIR"][f"CH_{num}"]
        if reg_en.content_enum == SvdEnum.ENABLED and reg_dir.content_enum != dir_enum:
            raise UicrError(
                f"Conflicting link direction for DPPIC 0x{addr:08x}: "
                f"{reg_dir.content_enum.lower()}, {dir_enum.lower()}"
            )
        self._map.reg_set(reg_en, SvdEnum.ENABLED, num_src)
        self._map.reg_set(reg_dir, dir_enum, direction_src)

        return self

    def add_ipct_channel(
        self,
        ipct_addr: int | Value[int],
        channel: Feature,
    ) -> Uicr:
        """Add an owned global domain IPCT channel for the local domain.

        :param ipct_addr: base address of the IPCT peripheral.
        :param channel: IPCT channel.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(ipct_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["IPCT"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(channel.num)
        self._map.reg_set(reg["CH"]["OWN"][f"CH_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(channel.secure)
        self._map.reg_set(
            reg["CH"]["SECURE"][f"CH_{num}"],
            _resolve_secure(reg["CH"]["SECURE"][f"CH_{num}"], secure),
            secure_src,
        )

        return self

    def add_ipct_interrupt(
        self,
        ipct_addr: int | Value[int],
        interrupt: Feature,
    ) -> Uicr:
        """Add an owned global domain IPCT interrupt for the local domain.

        :param ipct_addr: base address of the IPCT peripheral.
        :param interrupt: IPCT interrupt.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(ipct_addr)
        secure_addr = secure_address_get(addr)
        _, reg = _alloc_instanced_reg(self._uicr["IPCT"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(interrupt.num)
        self._map.reg_set(reg["INTERRUPT"]["OWN"][f"INT_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(interrupt.secure)
        self._map.reg_set(
            reg["INTERRUPT"]["SECURE"][f"INT_{num}"],
            _resolve_secure(reg["INTERRUPT"]["SECURE"][f"INT_{num}"], secure),
            secure_src,
        )

        return self

    def add_ipc_link(
        self,
        link: IpcLink,
    ) -> Uicr:
        """Link two IPCT channels together.

        :param link: IPCT channel link description.
        :returns: builder instance.
        """
        log.trace_function()

        source_domain, source_domain_src = _value_unpack(link.source_domain)
        source_num, source_num_src = _value_unpack(link.source_channel_num)
        sink_domain, sink_domain_src = _value_unpack(link.sink_domain)
        sink_num, sink_num_src = _value_unpack(link.sink_channel_num)

        for reg in self._uicr["IPCMAP"]:
            if not reg.modified or (
                reg["DOMAINIDSOURCE"].content,
                reg["IPCTCHSOURCE"].content,
                reg["DOMAINIDSINK"].content,
                reg["IPCTCHSINK"].content,
            ) == (source_domain, source_num, sink_domain, sink_num):
                self._map.reg_set(
                    reg["DOMAINIDSOURCE"], source_domain, source_domain_src
                )
                self._map.reg_set(reg["IPCTCHSOURCE"], source_num, source_num_src)
                self._map.reg_set(reg["DOMAINIDSINK"], sink_domain, sink_domain_src)
                self._map.reg_set(reg["IPCTCHSINK"], sink_num, sink_num_src)
                break
        else:
            raise UicrError(
                f"Too many entries in IPCMAP (max: {len(self._uicr['IPCMAP'])})"
            )

        return self

    def add_gpio_pin(self, gpio_addr: int | Value[int], pin: Pin) -> Uicr:
        """Add and configure an owned GPIO pin for the local domain.

        :param gpio_addr: base address of the GPIO peripheral.
        :param pin: GPIO pin description.
        :returns: builder instance.
        """
        log.trace_function()

        addr, addr_src = _value_unpack(gpio_addr)
        secure_addr = secure_address_get(addr)
        i, reg = _alloc_instanced_reg(self._uicr["GPIO"], secure_addr)

        if not reg["INSTANCE"].modified:
            self._map.reg_set(reg["INSTANCE"]["ADDRESS"], secure_addr, addr_src)

        num, num_src = _value_unpack(pin.num)
        self._map.reg_set(reg["OWN"][f"PIN_{num}"], SvdEnum.OWN, num_src)
        secure, secure_src = _value_unpack(pin.secure)
        self._map.reg_set(
            reg["SECURE"][f"PIN_{num}"],
            _resolve_secure(reg["SECURE"][f"PIN_{num}"], secure),
            secure_src,
        )
        ctrlsel, ctrlsel_src = _value_unpack(pin.ctrlsel)
        reg_ctrlsel = self._uicr["GPIO_PIN"][i]["CTRLSEL"][num]["CTRLSEL"]
        if reg_ctrlsel.modified and not (
            reg_ctrlsel.content == CTRLSEL_DEFAULT
            or ctrlsel == CTRLSEL_DEFAULT
            or ctrlsel == reg_ctrlsel.content
        ):
            raise UicrError(
                f"Conflicting CTRLSEL for GPIO 0x{addr:08x} pin {num}: "
                f"{reg_ctrlsel.content}, {ctrlsel}"
            )

        if not reg_ctrlsel.modified or ctrlsel != CTRLSEL_DEFAULT:
            self._map.reg_set(reg_ctrlsel, ctrlsel, ctrlsel_src)

        return self

    def set_secure_mailbox(self, mailbox: Mailbox) -> Uicr:
        """Configure the secure mode SSF IPC buffers for the local domain.

        :param mailbox: IPC buffer configuration.
        :returns: builder instance.
        """
        log.trace_function()

        reg = self._uicr["MAILBOX"]["SECURE"]

        self._map.reg_set(
            reg["TX_CONFIG0"]["ADDRESS"],
            *_value_unpack(mailbox.tx_address),
            lsb_truncate=True,
        )
        self._map.reg_set(
            reg["TX_CONFIG1"]["SIZE"],
            *_value_unpack(mailbox.tx_size),
            lsb_truncate=True,
        )
        self._map.reg_set(
            reg["RX_CONFIG0"]["ADDRESS"],
            *_value_unpack(mailbox.rx_address),
            lsb_truncate=True,
        )
        self._map.reg_set(
            reg["RX_CONFIG1"]["SIZE"],
            *_value_unpack(mailbox.rx_size),
            lsb_truncate=True,
        )

        return self

    def add_trace_link(self, link: Trace) -> Uicr:
        """Link a TDD trace source to a trace sink.

        :param link: trace link description.
        :returns: builder instance.
        """
        log.trace_function()

        sink, sink_src = _value_unpack(link.sink)
        if sink == Trace.Sink.ETB:
            reg = self._uicr["TRACE"]["ETBSINK"]["SOURCES"]
        elif sink == Trace.Sink.TPIU:
            reg = self._uicr["TRACE"]["TPIUSINK"]["SOURCES"]
        elif sink == Trace.Sink.ETR:
            reg = self._uicr["TRACE"]["ETRSINK"]["SOURCES"]
        else:
            raise ValueError(f"Invalid sink {link.sink}")

        source, source_src = _value_unpack(link.source)
        processor, processor_src = _value_unpack(link.processor)
        match (source, processor):
            case (Trace.Source.STM, ProcessorID.PPR):
                reg = reg["STMPPR"]
            case (Trace.Source.STM, ProcessorID.FLPR):
                reg = reg["STMFLPR"]
            case (Trace.Source.STM, ProcessorID.BBPR):
                reg = reg["STMBBPR"]
            case (Trace.Source.STM, _):
                reg = reg["STMMAINCORE"]
            case (Trace.Source.STM_HW_EVENTS, _):
                reg = reg["STMHWEVENTS"]
            case (Trace.Source.ETM, _):
                reg = reg["ETMMAINCORE"]
            case (Trace.Source.STM_LOCAL, _):
                raise NotImplementedError(
                    "Local STM trace links are not yet suppported"
                )
            case _:
                raise ValueError(f"Unrecognized trace link config: {link}")

        src_combined = []
        if processor_src is not None:
            src_combined.append(processor_src)
        if source_src is not None:
            src_combined.append(source_src)
        if sink_src is not None:
            src_combined.append(sink_src)

        self._map.reg_set(
            reg, SvdEnum.REQUESTED, src_combined if src_combined else None
        )

        return self

    def set_trace_port_config(self, config: TracePortConfig) -> Uicr:
        """Configure the trace port.

        :param config: trace port configuration.
        :returns: builder instance.
        """
        log.trace_function()

        self._map.reg_set(
            self._uicr["TRACE"]["PORTCONFIG"]["PORTCONFIG"],
            *_value_unpack(config.speed),
        )
        return self

    def set_etr_buffer(self, buffer: EtrBuffer) -> Uicr:
        """Configure the ETR buffer.

        :param buffer: ETR buffer configuration.
        :returns: builder instance.
        """
        log.trace_function()

        reg = self._uicr["TRACE"]["ETRBUF"]
        self._map.reg_set(
            reg["CONFIG0"]["ADDRESS"], *_value_unpack(buffer.address), lsb_truncate=True
        )
        self._map.reg_set(
            reg["CONFIG1"]["SIZE"], *_value_unpack(buffer.size), lsb_truncate=True
        )
        return self

    def pretty_print(self, *, verbose: bool = False) -> str:
        """Format the UICR as a human readable string.

        :param verbose: include exhaustive details about the UICR.
        :returns: human readable UICR
        """
        lines = []

        lines.append(self._pp_header(verbose=verbose))
        if vtor_str := self._pp_vtor(verbose=verbose):
            lines.append(_indent(vtor_str, 1))
        if memory_str := self._pp_memory(verbose=verbose):
            lines.append(_indent(memory_str, 1))
        if mailbox_str := self._pp_mailbox(verbose=verbose):
            lines.append(_indent(mailbox_str, 1))
        if periph_str := self._pp_periph(verbose=verbose):
            lines.append(_indent(periph_str, 1))
        if gpio_str := self._pp_gpio(verbose=verbose):
            lines.append(_indent(gpio_str, 1))
        if gpiote_str := self._pp_gpiote(verbose=verbose):
            lines.append(_indent(gpiote_str, 1))
        if grtc_str := self._pp_grtc(verbose=verbose):
            lines.append(_indent(grtc_str, 1))
        if dppi_str := self._pp_dppi(verbose=verbose):
            lines.append(_indent(dppi_str, 1))
        if ipct_str := self._pp_ipct(verbose=verbose):
            lines.append(_indent(ipct_str, 1))
        if ipcmap_str := self._pp_ipcmap(verbose=verbose):
            lines.append(_indent(ipcmap_str, 1))
        if trace_str := self._pp_trace(verbose=verbose):
            lines.append(_indent(trace_str, 1))

        return "\n".join(lines)

    def _pp_header(self, *, verbose: bool = False) -> str:
        uicr_header = self._uicr["HEADER"]
        version = (
            f"{uicr_header['VERSION']['MAJOR'].content}"
            f".{uicr_header['VERSION']['MINOR'].content}"
            f".{uicr_header['VERSION']['PATCH'].content}"
            f"+{uicr_header['VERSION']['TWEAK'].content}"
        )
        product_code = ProductCode(
            part_number=uicr_header["PARTNO"]["PARTNO"].content,
            revision=uicr_header["HWREVISION"]["HWREVISION"].content,
        )
        for product in Product:
            if product.product_code == product_code:
                product_name = product.name
                break
        else:
            product_name = f"Part no. 0x{product_code.part_number:x}, Rev. 0x{product_code.revision:x}"
        domain = uicr_header["DOMAIN"]["DOMAIN"].content
        try:
            domain_name = f"{DomainID(domain).name}"
        except Exception:
            domain_name = f"domain {domain}"

        return f"{product_name} {domain_name} UICR v{version}"

    def _pp_vtor(self, *, verbose: bool = False) -> str:
        lines = []

        if self._uicr["INITSVTOR"].modified:
            svtor = f"S: 0x{self._uicr['INITSVTOR']['INITSVTOR'].content:08x}"
        else:
            svtor = ""

        if self._uicr["INITNSVTOR"].modified:
            nsvtor = f"NS: 0x{self._uicr['INITNSVTOR']['INITNSVTOR'].content:08x}"
        else:
            nsvtor = ""

        if svtor or nsvtor:
            lines.append("Initial VTORs:")
            if svtor:
                lines.append(_indent(svtor, 1))
            if nsvtor:
                lines.append(_indent(nsvtor, 1))

        return "\n".join(lines)

    def _pp_memory(self, *, verbose: bool = False) -> str:
        lines = []
        mem_entries = []

        for i, uicr_mem in enumerate(self._uicr["MEM"]):
            if not uicr_mem["CONFIG0"].modified:
                break

            mem_address_val = content_get_and_extend(uicr_mem["CONFIG0"]["ADDRESS"])
            mem_size_val = content_get_and_extend(uicr_mem["CONFIG1"]["SIZE"])
            mem_address = f"0x{mem_address_val:x}"
            mem_size = f"0x{mem_size_val:x}"
            mem_address_end = f"0x{mem_address_val + mem_size_val:x}"
            mem_access_owner_val = uicr_mem["CONFIG1"]["OWNERID"].content
            try:
                mem_access_owner = OwnerID(mem_access_owner_val).name
            except Exception:
                mem_access_owner = str(mem_access_owner_val)
            read_str = (
                "R"
                if uicr_mem["CONFIG0"]["READ"].content_enum == SvdEnum.ALLOWED
                else ""
            )
            write_str = (
                "W"
                if uicr_mem["CONFIG0"]["WRITE"].content_enum == SvdEnum.ALLOWED
                else ""
            )
            execute_str = (
                "X"
                if uicr_mem["CONFIG0"]["EXECUTE"].content_enum == SvdEnum.ALLOWED
                else ""
            )
            secure_str = (
                "S"
                if uicr_mem["CONFIG0"]["SECURE"].content_enum == SvdEnum.SECURE
                else ""
            )
            mem_access = (
                f"{mem_access_owner}:{read_str}{write_str}{execute_str}{secure_str}"
            )

            for memory in mem_entries:
                if memory["address"] == mem_address and memory["size"] == mem_size:
                    memory["access"].append((i, mem_access))
                    break
            else:
                config_mem = {
                    "address": mem_address,
                    "address_end": mem_address_end,
                    "size": mem_size,
                    "access": [(i, mem_access)],
                }
                mem_entries.append(config_mem)

        if mem_entries:
            lines.append("Owned Memory:")
            for entry in mem_entries:
                numbers = ", ".join((f"#{i}" for i, _ in entry["access"]))
                access = ", ".join((a for _, a in entry["access"]))
                lines.append(
                    _indent(
                        f"{numbers}: [{entry['address']} - {entry['address_end']}) "
                        f"({entry['size']}) [{access}]",
                        1,
                    )
                )

        return "\n".join(lines)

    def _pp_mailbox(self, *, verbose: bool = False) -> str:
        lines = []
        uicr_sec_mailbox = self._uicr["MAILBOX"]["SECURE"]
        if uicr_sec_mailbox["TX_CONFIG0"].modified:
            tx_addr_val = content_get_and_extend(
                uicr_sec_mailbox["TX_CONFIG0"]["ADDRESS"]
            )
            tx_size_val = content_get_and_extend(uicr_sec_mailbox["TX_CONFIG1"]["SIZE"])
            rx_addr_val = content_get_and_extend(
                uicr_sec_mailbox["RX_CONFIG0"]["ADDRESS"]
            )
            rx_size_val = content_get_and_extend(uicr_sec_mailbox["RX_CONFIG1"]["SIZE"])
            lines.append("Secure SSF buffer:")
            lines.append(
                _indent(
                    f"TX: [0x{tx_addr_val:08x} - 0x{tx_addr_val + tx_size_val:08x}) "
                    f"(0x{tx_size_val:x})",
                    1,
                )
            )
            lines.append(
                _indent(
                    f"RX: [0x{rx_addr_val:08x} - 0x{rx_addr_val + rx_size_val:08x}) "
                    f"(0x{rx_size_val:x})",
                    1,
                )
            )
        return "\n".join(lines)

    def _pp_periph(self, *, verbose: bool = False) -> str:
        lines = []

        if self._uicr["PERIPH"][0]["CONFIG"].modified:
            lines.append("Owned peripherals:")

        for i, uicr_periph in enumerate(self._uicr["PERIPH"]):
            if not uicr_periph["CONFIG"].modified:
                break

            address = (
                f"0x{content_get_and_extend(uicr_periph['CONFIG']['ADDRESS']):08x}"
            )
            irq_processor_val = uicr_periph["CONFIG"]["PROCESSOR"].content
            try:
                irq_processor = ProcessorID(irq_processor_val).name
            except Exception:
                irq_processor = str(irq_processor_val)
            secure = (
                "S"
                if uicr_periph["CONFIG"]["SECURE"].content_enum == SvdEnum.SECURE
                else "NS"
            )
            dma_secure = (
                "DMA-S"
                if uicr_periph["CONFIG"]["DMASEC"].content_enum == SvdEnum.SECURE
                else "DMA-NS"
            )
            lines.append(
                _indent(
                    f"#{i}: {address} [{secure}, {dma_secure}, IRQ->{irq_processor}]", 1
                )
            )

        return "\n".join(lines)

    def _pp_gpiote(self, *, verbose: bool = False) -> str:
        lines = []

        for i, uicr_gpiote in enumerate(self._uicr["GPIOTE"]):
            if not uicr_gpiote["INSTANCE"].modified:
                break

            address = (
                f"0x{content_get_and_extend(uicr_gpiote['INSTANCE']['ADDRESS']):08x}"
            )
            lines.append(f"GPIOTE#{i} @ {address}")

            owned_channels = _pretty_print_owned_common(
                uicr_gpiote["CH"]["OWN"], uicr_gpiote["CH"]["SECURE"]
            )
            lines.append(_indent(f"owned channels: {owned_channels}", 1))

        return "\n".join(lines)

    def _pp_ipct(self, *, verbose: bool = False) -> str:
        lines = []

        for i, uicr_ipct in enumerate(self._uicr["IPCT"]):
            if not uicr_ipct["INSTANCE"].modified:
                break

            address = (
                f"0x{content_get_and_extend(uicr_ipct['INSTANCE']['ADDRESS']):08x}"
            )
            lines.append(f"IPCT#{i} @ {address}")

            if uicr_ipct["CH"]["OWN"].modified:
                owned_channels = _pretty_print_owned_common(
                    uicr_ipct["CH"]["OWN"], uicr_ipct["CH"]["SECURE"]
                )
                lines.append(_indent(f"owned channels: {owned_channels}", 1))

            if uicr_ipct["INTERRUPT"]["OWN"].modified:
                owned_interrupts = _pretty_print_owned_common(
                    uicr_ipct["INTERRUPT"]["OWN"],
                    uicr_ipct["INTERRUPT"]["SECURE"],
                )
                lines.append(_indent(f"owned interrupts: {owned_interrupts}", 1))

        return "\n".join(lines)

    def _pp_ipcmap(self, *, verbose: bool = False) -> str:
        lines = []

        if self._uicr["IPCMAP"][0].modified:
            lines.append("IPC mapping:")

        for i, uicr_ipcmap in enumerate(self._uicr["IPCMAP"]):
            if not uicr_ipcmap.modified:
                break

            source_domain_val = uicr_ipcmap["DOMAINIDSOURCE"].content
            try:
                source_domain = DomainID(source_domain_val).name
            except Exception:
                source_domain = str(source_domain_val)
            source_channel_num = uicr_ipcmap["IPCTCHSOURCE"].content

            sink_domain_val = uicr_ipcmap["DOMAINIDSINK"].content
            try:
                sink_domain = DomainID(sink_domain_val).name
            except Exception:
                sink_domain = str(sink_domain_val)
            sink_channel_num = uicr_ipcmap["IPCTCHSINK"].content

            lines.append(
                _indent(
                    f"#{i}: "
                    f"{source_domain} ch{source_channel_num} -> {sink_domain} ch{sink_channel_num}",
                    1,
                )
            )

        return "\n".join(lines)

    def _pp_dppi(self, *, verbose: bool = False) -> str:
        lines = []

        for i, uicr_dppi in enumerate(self._uicr["DPPI"]):
            if not uicr_dppi["INSTANCE"].modified:
                break

            address = (
                f"0x{content_get_and_extend(uicr_dppi['INSTANCE']['ADDRESS']):08x}"
            )
            lines.append(f"DPPIC#{i} @ {address}")

            if uicr_dppi["CH"]["OWN"].modified:
                owned_channels = _pretty_print_owned_common(
                    uicr_dppi["CH"]["OWN"], uicr_dppi["CH"]["SECURE"]
                )
                lines.append(_indent(f"owned channels: {owned_channels}", 1))

            if uicr_dppi["CHG"]["OWN"].modified:
                owned_channel_groups = _pretty_print_owned_common(
                    uicr_dppi["CHG"]["OWN"], uicr_dppi["CHG"]["SECURE"]
                )
                lines.append(
                    _indent(f"owned channel groups: {owned_channel_groups}", 1)
                )

            if uicr_dppi["CH"]["LINK"]["EN"].modified:
                source_links = []
                sink_links = []

                enabled = uicr_dppi["CH"]["LINK"]["EN"]
                direction = uicr_dppi["CH"]["LINK"]["DIR"]

                for i, (en_i, dir_i) in enumerate(
                    zip(enabled.values(), direction.values())
                ):
                    if en_i.content_enum not in (SvdEnum.LINKED, SvdEnum.ENABLED):
                        continue
                    if dir_i.content_enum == SvdEnum.SOURCE:
                        source_links.append(str(i))
                    else:
                        sink_links.append(str(i))

                if source_links:
                    lines.append(
                        _indent(f"source channels: {', '.join(source_links)}", 1)
                    )
                if sink_links:
                    lines.append(_indent(f"sink channels: {', '.join(sink_links)}", 1))

        return "\n".join(lines)

    def _pp_grtc(self, *, verbose: bool = False) -> str:
        lines = []

        if self._uicr["GRTC"]["CC"]["OWN"].modified:
            lines.append("GRTC:")
            owned_channels = _pretty_print_owned_common(
                self._uicr["GRTC"]["CC"]["OWN"],
                self._uicr["GRTC"]["CC"]["SECURE"],
            )
            lines.append(_indent(f"owned channels: {owned_channels}", 1))

        return "\n".join(lines)

    def _pp_gpio(self, *, verbose: bool = False) -> str:
        lines = []

        for i, uicr_gpio in enumerate(self._uicr["GPIO"]):
            if not uicr_gpio["INSTANCE"].modified:
                continue

            address = (
                f"0x{content_get_and_extend(uicr_gpio['INSTANCE']['ADDRESS']):08x}"
            )
            lines.append(f"GPIO#{i} @ {address}")

            entries = []
            for j, (own, sec) in enumerate(
                zip(uicr_gpio["OWN"].values(), uicr_gpio["SECURE"].values())
            ):
                owned = not own.content
                if not owned:
                    continue

                secure = "S" if sec.content_enum == SvdEnum.SECURE else "NS"
                uicr_ctrlsel = self._uicr["GPIO_PIN"][i]["CTRLSEL"][j]
                if uicr_ctrlsel.modified and uicr_ctrlsel.content != 0:
                    pin_ctrlsel = f" [ctrlsel={uicr_ctrlsel.content}]"
                else:
                    pin_ctrlsel = ""

                entries.append(f"{j}:{secure}{pin_ctrlsel}")

            owned_pins = ", ".join(entries)
            lines.append(_indent(f"owned pins: {owned_pins}", 1))

        return "\n".join(lines)

    def _pp_trace(self, *, verbose: bool = False) -> str:
        default_processor = ProcessorID.from_domain(
            self._uicr["HEADER"]["DOMAIN"]["DOMAIN"].content
        )
        uicr_trace = self._uicr["TRACE"]
        lines = []

        links = []
        for uicr_trace_sources, sink in [
            (uicr_trace["ETBSINK"]["SOURCES"], "etb"),
            (uicr_trace["TPIUSINK"]["SOURCES"], "tpiu"),
            (uicr_trace["ETRSINK"]["SOURCES"], "etr"),
        ]:
            for field, source, processor in [
                ("STMMAINCORE", "stm", default_processor),
                ("ETMMAINCORE", "etm", default_processor),
                ("STMPPR", "stm", ProcessorID.PPR),
                ("STMFLPR", "stm", ProcessorID.FLPR),
                ("STMBBPR", "stm", ProcessorID.BBPR),
                ("STMHWEVENTS", "stm-hw-events", default_processor),
                # TODO: 92 fields
            ]:
                if uicr_trace_sources[field].content_enum == SvdEnum.REQUESTED:
                    links.append(f"{source}:{processor.name} -> {sink}")

        if uicr_trace["PORTCONFIG"].modified:
            port_speed = str(uicr_trace["PORTCONFIG"]["PORTCONFIG"].content)
            port_config = f"trace port speed: {port_speed}"
        else:
            port_config = ""

        if uicr_trace["ETRBUF"]["CONFIG0"].modified:
            addr_val = content_get_and_extend(
                uicr_trace["ETRBUF"]["CONFIG0"]["ADDRESS"]
            )
            size_val = content_get_and_extend(uicr_trace["ETRBUF"]["CONFIG1"]["SIZE"])
            etr_buffer = f"etr buffer: [0x{addr_val:08x} - 0x{addr_val + size_val:08x}) ({size_val})"
        else:
            etr_buffer = ""

        if links or port_config or etr_buffer:
            lines.append("Trace:")

            if links:
                lines.append(_indent("trace links", 1))
                for link in links:
                    lines.append(_indent(link, 2))

            if etr_buffer:
                lines.append(_indent(etr_buffer, 1))

            if port_config:
                lines.append(_indent(port_config, 1))

        return "\n".join(lines)


def _alloc_instanced_reg(array: svd.Array, address: int) -> tuple[int, svd.Struct]:
    for i, reg in enumerate(array):
        if (
            not reg["INSTANCE"].modified
            or content_get_and_extend(reg["INSTANCE"]["ADDRESS"]) == address
        ):
            return i, reg

    raise UicrError(f"Too many entries in {array.path} (max: {len(array)})")


def _resolve_secure(existing: svd.Field, new: bool) -> str:
    if not new or existing.content_enum == SvdEnum.NONSECURE:
        return SvdEnum.NONSECURE
    return SvdEnum.SECURE


def _indent(text: str, level: int = 0) -> str:
    return textwrap.indent(text, prefix="  " * level)


def _pretty_print_owned_common(reg_own: svd.Register, reg_sec: svd.Register) -> str:
    entries = []

    for i, (own, sec) in enumerate(zip(reg_own.values(), reg_sec.values())):
        owned = not own.content
        if not owned:
            continue

        secure = "S" if sec.content_enum == SvdEnum.SECURE else "NS"
        entries.append(f"{i}:{secure}")

    return ", ".join(entries)



# Copyright (c) 2022 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Optional

from nrfregtool import _PACKAGE_NAME
from nrfregtool._cli import cli


def package_main(prog: Optional[str] = None):
    """
    Main entry point for the tool when installed as a package.
    """
    cli(prog=prog)


if __name__ == "__main__":
    package_main(prog=_PACKAGE_NAME)



# Copyright (c) 2023 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import argparse
import sys
from importlib.util import find_spec
from pathlib import Path
from textwrap import dedent
from typing import Any, Optional

import nrfregtool
from nrfregtool import RegtoolLogger, log
from nrfregtool.platform import Product


class CliError(RuntimeError):
    """An error caused by improper usage of the CLI."""

    ...


def cli(prog: Optional[str] = None, args_override: Optional[list[str]] = None) -> None:
    """Top level CLI entrypoint"""
    parser = make_arg_parser(prog)
    args = parser.parse_args(args_override)

    log_level = {
        0: RegtoolLogger.WARNING,
        1: RegtoolLogger.INFO,
        2: RegtoolLogger.DEBUG,
        3: RegtoolLogger.TRACE,
    }.get(args.verbose, RegtoolLogger.TRACE)
    log.setLevel(log_level)

    if not hasattr(args, "_command"):
        parser.print_help()
        sys.exit(2)

    kwargs = dict(vars(args))
    try:
        if args._command == "uicr-compile":
            cmd_uicr_compile(**kwargs)
        else:
            raise NotImplementedError(f"Unhandled command {args._command}")
    except CliError as e:
        log.error(str(e))
        sys.exit(2)
    except Exception as e:
        log.exception(e)
        sys.exit(2)

    sys.exit(0)


def make_arg_parser(prog: Optional[str] = None) -> argparse.ArgumentParser:
    top = argparse.ArgumentParser(
        prog=prog,
        description=dedent(
            """\
            CLI for generating UICR blobs for Haltium products.

            User Information Configuration Registers (UICR), in the context of certain
            Nordic SoCs, are used by Secure Domain Firmware (SDFW) to configure system
            resources, like memory and peripherals. When used in singular, a UICR refers
            to a complete set of registers for one domain.

            Use the -h or --help option with each command to see more information about
            them and their individual options/arguments.
            """
        ),
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    top.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        help="Print verbose output such as debug information.",
    )
    top.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s, version {nrfregtool.__version__}",
        help="Print program version",
    )

    sub = top.add_subparsers(title="subcommands")
    comp = sub.add_parser(
        "uicr-compile",
        help="Compile UICR from a devicetree",
        description=dedent(
            """\
            This command compiles a UICR binary blob from devicetree input, which should
            indicate the system resource configuration required for a single domain.

            For example, on nRF54H20:

                PYTHONPATH=zephyr/scripts/dts/python-devicetree/src \\
                nrf-regtool uicr-compile \\
                -e _build/zephyr/edt.pickle \\
                -o uicr.hex \\
                -P NRF54H20

            This command is intended for use as part of the Zephyr RTOS application
            development workflow. nrf-regtool depends on edtlib, a devicetree parsing
            library provided by the Zephyr Project, and the input file would usually be
            the edt.pickle artifact from a Zephyr build system. Using the pickle input
            requires a matching revision of edtlib from the Zephyr repository.
            """
        ),
        allow_abbrev=False,
        exit_on_error=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    comp.set_defaults(_command="uicr-compile")
    comp_in = comp.add_argument_group("Input file")
    comp_in.add_argument(
        "-e",
        "--edt-pickle-file",
        type=Path,
        required=True,
        help=(
            "Pickled edtlib.EDT object with parsed devicetree contents,\n"
            "containing information that is relevant for a domain's UICR. "
        ),
    )
    comp_in_opt = comp.add_argument_group("Input options")
    comp_in_opt.add_argument(
        "-P",
        "--product-name",
        required=True,
        help=(
            "Name of the Nordic SoC or SiP for which to compile the UICR.\n"
            f"Supported values are: {', '.join(x.name for x in Product)}.\n\n"
        ),
    )
    comp_out = comp.add_argument_group("Output files")
    comp_out.add_argument(
        "-o",
        "--output-file",
        required=True,
        type=Path,
        help="Generated output file of the encoded UICR in intel hex format.",
    )
    comp_out.add_argument(
        "--output-debug-file",
        type=Path,
        help="Debug information output in JSON format.",
    )

    return top


def cmd_uicr_compile(
    edt_pickle_file: Path,
    product_name: str,
    output_file: Path,
    output_debug_file: Optional[Path] = None,
    **kwargs: Any,
):
    if find_spec("devicetree") is None:
        raise CliError(
            "Devicetree support is not available outside the zephyr build system"
        )
    try:
        product = Product(product_name)
    except ValueError:
        raise CliError(
            f"Unknown product name {product_name}; must be one of {[x.name for x in Product]}"
        )

    from nrfregtool.zephyr.uicr_compile import uicr_from_devicetree

    uicr_from_devicetree(edt_pickle_file, product, output_file, output_debug_file)
